# Copyright 2025 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#
from http import HTTPStatus
import json
from typing import Any
import httpx
import pytest
import respx

from planet import FeaturesClient, Session
from planet.auth import APIKeyAuth
from planet.sync.features import FeaturesAPI

pytestmark = pytest.mark.anyio  # noqa

# Simulated host/path for testing purposes. Not a real subdomain.
TEST_URL = "http://test.planet.com/features/v1/ogc/my"

TEST_GEOM = {"type": "Polygon", "coordinates": [[[]]]}

TEST_FEAT = {"type": "Feature", "geometry": TEST_GEOM}

TEST_COLLECTION_1 = {
    "id": "collection1",
    "title": "Collection 1",
    "description": "test collection 1",
}

TEST_COLLECTION_2 = {
    "id": "collection2",
    "title": "Collection 2",
    "description": "test collection 2",
}

TEST_COLLECTION_LIST = [TEST_COLLECTION_1, TEST_COLLECTION_2]

# set up test clients
test_session = Session(auth=APIKeyAuth(key="test"))
cl_async = FeaturesClient(test_session, base_url=TEST_URL)
cl_sync = FeaturesAPI(test_session, base_url=TEST_URL)


class ExampleGeoInterface:

    @property
    def __geo_interface__(self) -> dict:
        return TEST_GEOM


def mock_response(url: str, json: Any, method: str = "get"):
    mock_resp = httpx.Response(HTTPStatus.OK, json=json)
    respx.request(method, url).return_value = mock_resp


def to_collection_model(collection: dict) -> dict:
    """create a collection model as it would appear in a Features API response"""
    id = collection.get("id")

    # define a couple of the longer values up front for formatting reasons
    link = f"https://api.planet.com/features/v0/ogc/my/collections/{id}/items"
    permissions = {"can_write": True, "shared": False, "is_owner": True}

    return {
        "id": id,
        "title": collection.get("title"),
        "description": collection.get("description"),
        "item_type": "feature",
        "extent": {
            "spatial": {
                "bbox": [[0, 0, 1, 1]]
            }
        },
        "links": [
            {
                "href": f"{TEST_URL}/collections/{id}",
                "rel": "self",
                "title": "This collection",
                "type": "application/json",
            },
            {
                "href": link,
                "rel": "features",
                "title": "Features",
                "type": "application/json",
            },
        ],
        "feature_count": 1,
        "area": 1,
        "title_property": "NAME_0",
        "description_property": "description",
        "permissions": permissions,
    }


def list_collections_response(collections: list[dict]) -> dict:
    """simulate a list collections response"""
    return {
        "numberMatched": len(collections),
        "links": [{
            "href": f"{TEST_URL}/collections",
            "rel": "self",
            "title": "This page of results",
        }],
        "collections": [
            to_collection_model(collection) for collection in collections
        ],
    }


def to_feature_model(id: str) -> dict:
    return {
        "type": "Feature",
        "id": id,
        "properties": {},
        "geometry": {
            "coordinates": [[
                [7.05322265625, 46.81509864599243],
                [7.580566406250001, 46.81509864599243],
                [7.580566406250001, 47.17477833929903],
                [7.05322265625, 47.17477833929903],
                [7.05322265625, 46.81509864599243],
            ]],
            "type": "Polygon",
        },
    }


def list_features_response(collection_id: str, num_features: int) -> dict:
    return {
        "type": "FeatureCollection",
        "numberMatched": num_features,
        "links": [{
            "href": f"https://api.planet.com/features/v0/ogc/my/collections/{collection_id}/items",
            "rel": "self",
            "title": "This page of results",
        }],
        "features": [to_feature_model(str(i)) for i in range(num_features)],
    }


@respx.mock
async def test_list_collections():

    collections_url = f"{TEST_URL}/collections"
    mock_response(
        collections_url,
        list_collections_response([TEST_COLLECTION_1, TEST_COLLECTION_2]),
    )

    def assertf(resp):
        assert resp[0]["id"] == "collection1"
        assert resp[1]["id"] == "collection2"

    assertf([col async for col in cl_async.list_collections()])
    assertf(list(cl_sync.list_collections()))


@respx.mock
async def test_get_collection():

    id = TEST_COLLECTION_1["id"]
    collection_url = f"{TEST_URL}/collections/{id}"

    mock_response(collection_url, to_collection_model(TEST_COLLECTION_1))

    def assertf(resp):
        assert resp["id"] == "collection1"
        assert resp["title"] == "Collection 1"

    assertf(await cl_async.get_collection(id))
    assertf(cl_sync.get_collection(id))


@respx.mock
async def test_create_collection():

    collection_url = f"{TEST_URL}/collections"

    mock_response(collection_url,
                  to_collection_model(TEST_COLLECTION_1),
                  method="post")

    def assertf(resp):
        # the return value is simply the id.
        assert resp == "collection1"

    params = {
        "title": TEST_COLLECTION_1["title"],
        "description": TEST_COLLECTION_1["description"],
    }

    assertf(await cl_async.create_collection(**params))
    assertf(cl_sync.create_collection(**params))

    req_body = json.loads(respx.calls[0].request.content)
    assert req_body["title"] == "Collection 1"
    assert req_body["description"] == "test collection 1"


@respx.mock
async def test_list_items():
    collection_id = "test"
    items_url = f"{TEST_URL}/collections/{collection_id}/items"

    mock_response(items_url,
                  list_features_response(collection_id, num_features=3))

    def assertf(resp):
        assert resp[0]["id"] == "0"
        assert resp[1]["id"] == "1"

    assertf([feat async for feat in cl_async.list_items(collection_id)])
    assertf(list(cl_sync.list_items(collection_id)))


@respx.mock
@pytest.mark.parametrize(
    "feature, expected_body",
    [
        (TEST_FEAT, TEST_GEOM),
        (TEST_GEOM, TEST_GEOM),
        (ExampleGeoInterface(), TEST_GEOM),
    ],
)
async def test_add_items(feature, expected_body):
    """test adding a feature with the SDK
    cases:
    * a geojson Feature
    * a geojson Geometry
    * an object that implements __geo_interface__
    """
    collection_id = "test"
    items_url = f"{TEST_URL}/collections/{collection_id}/items"

    # mock a feature ref return
    feat_resp = ["pl:features/my/test/test1"]
    mock_response(items_url, feat_resp, method="post")

    def assertf(resp):
        # check that mocked response is returned in full
        assert resp == feat_resp

    assertf(await cl_async.add_items(collection_id, feature))
    assertf(cl_sync.add_items(collection_id, feature))

    # check request body. In all test cases, the request body
    # should be a geojson Feature.
    req_body = json.loads(respx.calls[0].request.content)
    assert req_body["type"] == "Feature"
    assert req_body["geometry"] == expected_body



# Copyright 2020 Planet Labs, Inc.
# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import logging

import pytest

from planet import auth

LOGGER = logging.getLogger(__name__)


# skip the global mock of _SecretFile.read
# for this module
@pytest.fixture(autouse=True, scope='module')
def test_secretfile_read():
    return


@pytest.fixture
def secret_path(monkeypatch, tmp_path):
    secret_path = tmp_path / '.test'
    monkeypatch.setattr(auth, 'SECRET_FILE_PATH', secret_path)
    yield secret_path


def test_Auth_from_key():
    test_auth_env1 = auth.Auth.from_key('testkey')
    assert test_auth_env1.value == 'testkey'


def test_Auth_from_key_empty():
    with pytest.raises(auth.APIKeyAuthException):
        _ = auth.Auth.from_key('')


def test_Auth_from_file(secret_path):
    with open(secret_path, 'w') as fp:
        fp.write('{"key": "testvar"}')

    test_auth = auth.Auth.from_file()
    assert test_auth.value == 'testvar'


def test_Auth_from_file_doesnotexist(secret_path):
    with pytest.raises(auth.AuthException):
        _ = auth.Auth.from_file(secret_path)


def test_Auth_from_file_wrongformat(secret_path):
    with open(secret_path, 'w') as fp:
        fp.write('{"notkey": "testvar"}')

    with pytest.raises(auth.AuthException):
        _ = auth.Auth.from_file(secret_path)


def test_Auth_from_file_alternate(tmp_path):
    secret_path = str(tmp_path / '.test')
    with open(secret_path, 'w') as fp:
        fp.write('{"key": "testvar"}')

    test_auth = auth.Auth.from_file(secret_path)
    assert test_auth.value == 'testvar'


def test_Auth_from_env(monkeypatch):
    monkeypatch.setenv('PL_API_KEY', 'testkey')
    test_auth_env = auth.Auth.from_env()
    assert test_auth_env.value == 'testkey'


def test_Auth_from_env_failure(monkeypatch):
    monkeypatch.delenv('PL_API_KEY', raising=False)
    with pytest.raises(auth.AuthException):
        _ = auth.Auth.from_env()


def test_Auth_from_env_alternate_success(monkeypatch):
    alternate = 'OTHER_VAR'
    monkeypatch.setenv(alternate, 'testkey')
    monkeypatch.delenv('PL_API_KEY', raising=False)

    test_auth_env = auth.Auth.from_env(alternate)
    assert test_auth_env.value == 'testkey'


def test_Auth_from_env_alternate_doesnotexist(monkeypatch):
    alternate = 'OTHER_VAR'
    monkeypatch.delenv(alternate, raising=False)
    monkeypatch.delenv('PL_API_KEY', raising=False)

    with pytest.raises(auth.AuthException):
        _ = auth.Auth.from_env(alternate)


def test_Auth_from_login(monkeypatch):
    auth_data = 'authdata'

    def login(*args, **kwargs):
        return {'api_key': auth_data}

    monkeypatch.setattr(auth.AuthClient, 'login', login)

    test_auth = auth.Auth.from_login('email', 'pw')
    assert test_auth.value == auth_data


def test_Auth_store_doesnotexist(tmp_path):
    test_auth = auth.Auth.from_key('test')
    secret_path = str(tmp_path / '.test')
    test_auth.store(secret_path)

    with open(secret_path, 'r') as fp:
        assert json.loads(fp.read()) == {"key": "test"}


def test_Auth_store_exists(tmp_path):
    secret_path = str(tmp_path / '.test')

    with open(secret_path, 'w') as fp:
        fp.write('{"existing": "exists"}')

    test_auth = auth.Auth.from_key('test')
    test_auth.store(secret_path)

    with open(secret_path, 'r') as fp:
        assert json.loads(fp.read()) == {"key": "test", "existing": "exists"}


def test__SecretFile_permissions_doesnotexist(secret_path):
    """No exception is raised if the file doesn't exist"""
    auth._SecretFile(secret_path)


def test__SecretFile_permissions_incorrect(secret_path):
    """Incorrect permissions are fixed"""
    with open(secret_path, 'w') as fp:
        fp.write('{"existing": "exists"}')

    secret_path.chmod(0o666)

    auth._SecretFile(secret_path)
    assert secret_path.stat().st_mode & 0o777 == 0o600



# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Decorators for Click commands"""
import asyncio
from functools import wraps
from typing import Callable, Optional

import click

from planet import exceptions
from planet.cli.options import pretty


def command(group: click.Group,
            name: Optional[str] = None,
            extra_args: list[Callable] = []):
    """a decorator that adds common utilities/options to a click command

    usage:

    @command(features)  # pass a group created with @click.group
    def my_command():
        ...

    this single decorator replaces a list of decorators we would otherwise
    add to every function e.g.:

    @features.command
    @coro
    @translate_exceptions
    @click.pass_context
    def my_command():
        ...
    """

    # the decorators to add to the function **when the function is run**
    # (as opposed to when the function is registered as a click command)
    decorators = [
        coro,
        translate_exceptions,
        click.pass_context,
        pretty,
    ] + extra_args

    # since we want to use `command` as a function with an arg: `@command(group)`,
    # we need to create and return an "real" decorator that takes the function as its
    # arg.
    def decorator(f):

        # run any click-specific registration decorators
        for fn in decorators:
            f = fn(f)

        @wraps(f)
        def wrapper(*args, **kwargs):
            cmd = f

            # wrap cmd with all the default decorators
            for d in decorators:
                cmd = d(f)

            return cmd(*args, **kwargs)

        # register the whole thing as a Click command.
        # Doing this last (outside the wrapper) allows click to
        # pick up the options/arguments added to the command by e.g.
        # `@click.option()`
        return group.command(name=name)(wrapper)

    return decorator


# https://github.com/pallets/click/issues/85#issuecomment-503464628
def coro(func):
    """Wrap async functions so they can be run sync with Click.

    Parameters:
        func: a Click command function or wrapper around one.

    Returns:
        wrapper function
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        return asyncio.run(func(*args, **kwargs))

    return wrapper


def translate_exceptions(func):
    """Translate internal exceptions to ClickException.

    Parameters:
        func: a Click command function or wrapper around one.

    Returns:
        wrapper function

    Raises:
        ClickException
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except exceptions.AuthException:
            raise click.ClickException(
                'Auth information does not exist or is corrupted. Initialize '
                'with `planet auth init`.')
        except exceptions.PlanetError as ex:
            raise click.ClickException(str(ex))

    return wrapper



# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Test Orders CLI"""
import copy
import hashlib
from http import HTTPStatus
import json
from pathlib import Path
from unittest.mock import Mock

from click.testing import CliRunner
import httpx
import pytest
import respx

from planet.cli import cli

TEST_URL = 'http://MockNotRealURL/api/path'
TEST_DOWNLOAD_URL = f'{TEST_URL}/download'
TEST_ORDERS_URL = 'https://api.planet.com/compute/ops/orders/v2'

# NOTE: These tests use a lot of the same mocked responses as test_orders_api.


@pytest.fixture
def invoke():

    def _invoke(extra_args, runner=None):
        runner = runner or CliRunner()
        args = ['orders'] + extra_args
        return runner.invoke(cli.main, args=args)

    return _invoke


@pytest.fixture
def stac_json():
    return {'stac': {}}


@respx.mock
def test_cli_orders_list_basic(invoke, order_descriptions):
    next_page_url = TEST_ORDERS_URL + '/blob/?page_marker=IAmATest'
    order1, order2, order3 = order_descriptions
    page1_response = {
        "_links": {
            "_self": "string", "next": next_page_url
        },
        "orders": [order1, order2]
    }
    mock_resp1 = httpx.Response(HTTPStatus.OK, json=page1_response)
    respx.get(TEST_ORDERS_URL).return_value = mock_resp1

    page2_response = {"_links": {"_self": next_page_url}, "orders": [order3]}
    mock_resp2 = httpx.Response(HTTPStatus.OK, json=page2_response)
    respx.get(next_page_url).return_value = mock_resp2

    result = invoke(['list'])
    assert result.exit_code == 0
    sequence = '\n'.join([json.dumps(o) for o in [order1, order2, order3]])
    assert result.output == sequence + '\n'


@respx.mock
def test_cli_orders_list_empty(invoke):
    page1_response = {"_links": {"_self": "string"}, "orders": []}
    mock_resp = httpx.Response(HTTPStatus.OK, json=page1_response)
    respx.get(TEST_ORDERS_URL).return_value = mock_resp

    result = invoke(['list'])
    assert result.exit_code == 0
    assert result.output == ''


@respx.mock
def test_cli_orders_list_filtering_and_sorting(invoke, order_descriptions):
    list_url = TEST_ORDERS_URL + '?source_type=all&state=failed&name=my_order_xyz&name__contains=xyz&created_on=2018-02-12T00:00:00Z/..&last_modified=../2018-03-18T12:31:12Z&hosting=true&sort_by=name DESC'

    order1, order2, _ = order_descriptions

    page1_response = {
        "_links": {
            "_self": "string"
        }, "orders": [order1, order2]
    }
    mock_resp = httpx.Response(HTTPStatus.OK, json=page1_response)
    respx.get(list_url).return_value = mock_resp

    # if the value of each arg doesn't get sent as a url parameter,
    # the mock will fail and this test will fail
    result = invoke([
        'list',
        '--state',
        'failed',
        '--name',
        'my_order_xyz',
        '--name-contains',
        'xyz',
        '--created-on',
        '2018-02-12T00:00:00Z/..',
        '--last-modified',
        '../2018-03-18T12:31:12Z',
        '--hosting',
        'true',
        '--sort-by',
        'name DESC'
    ])
    assert result.exit_code == 0
    sequence = '\n'.join([json.dumps(o) for o in [order1, order2]])
    assert result.output == sequence + '\n'


@respx.mock
@pytest.mark.parametrize("limit,limited_list_length", [(None, 100), (0, 102),
                                                       (1, 1)])
def test_cli_orders_list_limit(invoke,
                               order_descriptions,
                               limit,
                               limited_list_length):
    # Creating 102 (3x34) order descriptions
    long_order_descriptions = order_descriptions * 34

    all_orders = {}
    for x in range(1, len(long_order_descriptions) + 1):
        all_orders["order{0}".format(x)] = long_order_descriptions[x - 1]

    page1_response = {
        "_links": {
            "_self": "string"
        },
        "orders": [
            all_orders['order%s' % num]
            for num in range(1, limited_list_length + 1)
        ]
    }
    mock_resp = httpx.Response(HTTPStatus.OK, json=page1_response)

    # limiting is done within the client, no change to api call
    respx.get(TEST_ORDERS_URL).return_value = mock_resp

    result = invoke(['list', '--limit', limit])
    assert result.exit_code == 0
    count = len(result.output.strip().split('\n'))
    assert count == limited_list_length


@respx.mock
def test_cli_orders_list_pretty(invoke, monkeypatch, order_description):
    mock_echo_json = Mock()
    monkeypatch.setattr(cli.orders, 'echo_json', mock_echo_json)

    page1_response = {
        "_links": {
            "_self": "string"
        }, "orders": [order_description]
    }
    mock_resp = httpx.Response(HTTPStatus.OK, json=page1_response)
    respx.get(TEST_ORDERS_URL).return_value = mock_resp

    result = invoke(['list', '--pretty'])
    assert result.exit_code == 0
    mock_echo_json.assert_called_once_with(order_description, True)


# TODO: add tests for "get --pretty" (gh-491).
@respx.mock
def test_cli_orders_get(invoke, oid, order_description):
    get_url = f'{TEST_ORDERS_URL}/{oid}'
    mock_resp = httpx.Response(HTTPStatus.OK, json=order_description)
    respx.get(get_url).return_value = mock_resp

    result = invoke(['get', oid])
    assert result.exit_code == 0
    assert order_description == json.loads(result.output)


@respx.mock
def test_cli_orders_get_id_not_found(invoke, oid):
    get_url = f'{TEST_ORDERS_URL}/{oid}'
    error_json = {"message": "Error message"}
    mock_resp = httpx.Response(404, json=error_json)
    respx.get(get_url).return_value = mock_resp

    result = invoke(['get', oid])
    assert result.exit_code == 1
    assert 'Error: {"message":"Error message"}\n' == result.output


# TODO: add tests for "cancel --pretty" (gh-491).
@respx.mock
def test_cli_orders_cancel(invoke, oid, order_description):
    cancel_url = f'{TEST_ORDERS_URL}/{oid}'
    order_description['state'] = 'cancelled'
    mock_resp = httpx.Response(HTTPStatus.OK, json=order_description)
    respx.put(cancel_url).return_value = mock_resp

    result = invoke(['cancel', oid])
    assert result.exit_code == 0
    assert str(mock_resp.json()) + '\n' == result.output


@respx.mock
def test_cli_orders_cancel_id_not_found(invoke, oid):
    cancel_url = f'{TEST_ORDERS_URL}/{oid}'
    error_json = {"message": "Error message"}
    mock_resp = httpx.Response(404, json=error_json)
    respx.put(cancel_url).return_value = mock_resp

    result = invoke(['cancel', oid])
    assert result.exit_code == 1
    assert 'Error: {"message":"Error message"}\n' == result.output


# TODO: add tests for "wait --state" (gh-492) and "wait --pretty" (gh-491).
@respx.mock
def test_cli_orders_wait_default(invoke, order_description, oid):
    get_url = f'{TEST_ORDERS_URL}/{oid}'

    order_description2 = copy.deepcopy(order_description)
    order_description2['state'] = 'success'

    route = respx.get(get_url)
    route.side_effect = [
        httpx.Response(HTTPStatus.OK, json=order_description),
        httpx.Response(HTTPStatus.OK, json=order_description2)
    ]

    runner = CliRunner()
    result = invoke(['wait', '--delay', '0', oid], runner=runner)
    assert result.exit_code == 0
    assert result.output.endswith('success\n')


@respx.mock
def test_cli_orders_wait_max_attempts(invoke, order_description, oid):
    get_url = f'{TEST_ORDERS_URL}/{oid}'

    order_description2 = copy.deepcopy(order_description)
    order_description2['state'] = 'running'
    order_description3 = copy.deepcopy(order_description)
    order_description3['state'] = 'success'

    route = respx.get(get_url)
    route.side_effect = [httpx.Response(HTTPStatus.OK, json=order_description)]

    runner = CliRunner()
    result = invoke(['wait', '--delay', '0', '--max-attempts', '1', oid],
                    runner=runner)
    assert result.exit_code == 1
    assert result.output.endswith(
        'Error: Maximum number of attempts (1) reached.\n')


@pytest.fixture
def mock_download_response(oid, order_description):

    def _func():
        # Mock an HTTP response for download
        order_description['state'] = 'success'
        dl_url1 = TEST_DOWNLOAD_URL + '/1?token=IAmAToken'
        dl_url2 = TEST_DOWNLOAD_URL + '/2?token=IAmAnotherToken'
        dl_url3 = TEST_DOWNLOAD_URL + '/manifest'

        order_description['_links']['results'] = [{
            'location': dl_url1,
            'name': f'{oid}/itemtype/m1.json'
        }, {
            'location': dl_url2,
            'name': f'{oid}/itemtype/m2.json'
        }, {
            'location': dl_url3,
            'name': f'{oid}/manifest.json'
        }]  # yapf: disable

        get_url = f'{TEST_ORDERS_URL}/{oid}'
        mock_resp = httpx.Response(HTTPStatus.OK, json=order_description)
        respx.get(get_url).return_value = mock_resp

        mock_resp1 = httpx.Response(
            HTTPStatus.OK,
            json={'key': 'value'},
            headers={
                'Content-Type': 'application/json',
                'Content-Disposition': 'attachment; filename="m1.json"'
            })
        respx.get(dl_url1).return_value = mock_resp1

        mock_resp2 = httpx.Response(
            HTTPStatus.OK,
            json={'key2': 'value2'},
            headers={
                'Content-Type': 'application/json',
                'Content-Disposition': 'attachment; filename="m2.json"'
            })
        respx.get(dl_url2).return_value = mock_resp2

        m1_bytes = b'{"key":"value"}'
        m2_bytes = b'{"key2":"value2"}'
        manifest_data = {
            "name": "",
            "files": [
                {
                    "path": "itemtype/m1.json",
                    "digests": {
                        "md5": hashlib.md5(m1_bytes).hexdigest(),
                        "sha256": hashlib.sha256(m1_bytes).hexdigest()}
                }, {
                    "path": "itemtype/m2.json",
                    "digests": {
                        "md5": hashlib.md5(m2_bytes).hexdigest(),
                        "sha256": hashlib.sha256(m2_bytes).hexdigest()}
                }]
        }  # yapf: disable
        mock_resp3 = httpx.Response(
            HTTPStatus.OK,
            json=manifest_data,
            headers={
                'Content-Type': 'application/json',
                'Content-Disposition': 'attachment; filename="manifest.json"'
            })
        respx.get(dl_url3).return_value = mock_resp3

    return _func


@respx.mock
def test_cli_orders_download_default(invoke, mock_download_response, oid):
    mock_download_response()

    runner = CliRunner()
    with runner.isolated_filesystem() as folder:
        result = invoke(['download', oid], runner=runner)
        assert result.exit_code == 0

        # basic check of progress reporting
        assert 'm1.json' in result.output

        # Check that the files were downloaded and have the correct contents
        with open(Path(folder) / f'{oid}/itemtype/m1.json') as f:
            assert json.load(f) == {'key': 'value'}
        with open(Path(folder) / f'{oid}/itemtype/m2.json') as f:
            assert json.load(f) == {'key2': 'value2'}


@respx.mock
def test_cli_orders_download_checksum(invoke, mock_download_response, oid):
    """checksum is successful"""
    mock_download_response()

    runner = CliRunner()
    with runner.isolated_filesystem():
        result = invoke(['download', oid, '--checksum=MD5'], runner=runner)
        assert result.exit_code == 0


@respx.mock
def test_cli_orders_download_dest(invoke, mock_download_response, oid):
    mock_download_response()

    runner = CliRunner()
    with runner.isolated_filesystem() as folder:
        dest_dir = Path(folder) / 'foobar'
        dest_dir.mkdir()
        result = invoke(['download', '--directory', 'foobar', oid],
                        runner=runner)
        assert result.exit_code == 0

        # Check that the files were downloaded to the custom directory
        with open(dest_dir / f'{oid}/itemtype/m1.json') as f:
            assert json.load(f) == {'key': 'value'}

        with open(dest_dir / f'{oid}/itemtype/m2.json') as f:
            assert json.load(f) == {'key2': 'value2'}


@respx.mock
def test_cli_orders_download_overwrite(invoke,
                                       mock_download_response,
                                       oid,
                                       write_to_tmp_json_file):
    mock_download_response()

    runner = CliRunner()
    with runner.isolated_filesystem() as folder:
        filepath = Path(folder) / f'{oid}/itemtype/m1.json'
        filepath.parent.mkdir(parents=True)
        filepath.write_text(json.dumps({'foo': 'bar'}))

        # check the file doesn't get overwritten by default
        result = invoke(['download', oid], runner=runner)
        assert result.exit_code == 0

        with open(filepath, 'r') as f:
            assert json.load(f) == {'foo': 'bar'}

        # check the file gets overwritten
        result = invoke(['download', '--overwrite', oid], runner=runner)
        assert result.exit_code == 0

        with open(filepath, 'r') as f:
            assert json.load(f) == {'key': 'value'}


@respx.mock
def test_cli_orders_download_state(invoke, order_description, oid):
    get_url = f'{TEST_ORDERS_URL}/{oid}'

    order_description['state'] = 'running'
    mock_resp = httpx.Response(HTTPStatus.OK, json=order_description)
    respx.get(get_url).return_value = mock_resp

    runner = CliRunner()
    result = invoke(['download', oid], runner=runner)

    assert result.exit_code == 1
    assert 'order state (running) is not a final state.' in result.output


@respx.mock
def test_cli_orders_create_basic_success(invoke, order_description):
    mock_resp = httpx.Response(HTTPStatus.OK, json=order_description)
    respx.post(TEST_ORDERS_URL).return_value = mock_resp

    order_request = {
        "name": "test",
        "products": [{
            "item_ids": ["20220325_131639_20_2402"],
            "item_type": "PSScene",
            "product_bundle": "visual",
        }],
    }

    result = CliRunner().invoke(
        cli.main, ['orders', 'create', json.dumps(order_request)],
        catch_exceptions=True)

    assert result.exit_code == 0
    assert json.loads(
        result.output)['_links']['results'][0]['delivery'] == 'success'


@pytest.mark.parametrize(
    "id_string, expected_ids",
    [('20220325_131639_20_2402', ['20220325_131639_20_2402']),
     ('20220325_131639_20_2402,20230324_121730_43_2423',
      ['20220325_131639_20_2402', '20230324_121730_43_2423'])])
def test_cli_orders_request_basic_success(mock_bundles,
                                          expected_ids,
                                          id_string,
                                          invoke,
                                          stac_json):
    result = invoke([
        'request',
        '--item-type=PSScene',
        '--bundle=visual',
        '--name=test',
        id_string,
    ])
    assert not result.exception

    order_request = {
        "name": "test",
        "products": [{
            "item_ids": expected_ids,
            "item_type": "PSScene",
            "product_bundle": "visual"
        }],
        "metadata": stac_json
    }
    assert order_request == json.loads(result.output)


def test_cli_orders_request_item_type_invalid(mock_bundles, invoke):
    result = invoke([
        'request',
        '--item-type=invalid'
        '--bundle=visual',
        '--name=test',
        '20220325_131639_20_2402',
    ])
    assert result.exit_code == 2


def test_cli_orders_request_product_bundle_invalid(mock_bundles, invoke):
    result = invoke([
        'request',
        '--item-type=PSScene'
        '--bundle=invalid',
        '--name=test',
        '20220325_131639_20_2402',
    ])
    assert result.exit_code == 2


def test_cli_orders_request_product_bundle_incompatible(mock_bundles, invoke):
    result = invoke([
        'request',
        '--item-type=PSScene',
        '--bundle=analytic',
        '--name=test',
        '20220325_131639_20_2402',
    ])
    assert result.exit_code == 2


@pytest.mark.parametrize(
    "bundle, fallback_bundle",
    [
        ('analytic_8b_udm2', 'analytic_udm2'),
        ('analytic_8b_udm2', 'analytic_udm2,analytic_3b_udm2'),
    ])
def test_cli_orders_request_fallback_bundle_success(invoke,
                                                    bundle,
                                                    fallback_bundle):
    result = invoke([
        'request',
        '--item-type=PSScene',
        f'--bundle={bundle}',
        f'--fallback-bundle={fallback_bundle}',
        '--name=my order',
        '20250130_035211_69_2516'
    ])
    assert result.exit_code == 0


@pytest.mark.parametrize(
    "bundle, fallback_bundle",
    [
        ('analytic_8b_udm2', ''),
        ('analytic_8b_udm2', 'analytic_udm2;analytic_3b_udm2'),
        ('analytic_8b_udm2', 'analytic_udm2,analytic_3b_udm2,fake_bundle'),
    ])
def test_cli_orders_request_fallback_bundle_fail(invoke,
                                                 bundle,
                                                 fallback_bundle):
    result = invoke([
        'request',
        '--item-type=PSScene',
        f'--bundle={bundle}',
        f'--fallback-bundle={fallback_bundle}',
        '--name=my order',
        '20250130_035211_69_2516'
    ])
    assert result.exit_code == 2


def test_cli_orders_request_id_empty(mock_bundles, invoke):
    result = invoke([
        'request', '--item-type=PSScene', '--bundle=visual', '--name=test', ''
    ])
    assert result.exit_code == 2
    assert 'Entry cannot be an empty string.' in result.output


@pytest.mark.parametrize("geom_fixture",
                         [('geom_geojson'), ('feature_geojson'),
                          ('featurecollection_geojson')])
def test_cli_orders_request_clip_polygon(mock_bundles,
                                         geom_fixture,
                                         request,
                                         invoke,
                                         geom_geojson,
                                         stac_json):

    geom = request.getfixturevalue(geom_fixture)

    result = invoke([
        'request',
        '--item-type=PSScene',
        '--bundle=visual',
        '--name=test',
        '20220325_131639_20_2402',
        f'--clip={json.dumps(geom)}',
    ])
    assert result.exit_code == 0

    order_request = {
        "name": "test",
        "products": [{
            "item_ids": ["20220325_131639_20_2402"],
            "item_type": "PSScene",
            "product_bundle": "visual",
        }],
        "tools": [{
            'clip': {
                'aoi': geom_geojson
            }
        }],
        "metadata": stac_json
    }
    assert order_request == json.loads(result.output)


@pytest.mark.parametrize("geom_fixture", [('geom_reference')])
def test_cli_orders_request_clip_ref(mock_bundles,
                                     geom_fixture,
                                     request,
                                     invoke,
                                     stac_json):

    geom = request.getfixturevalue(geom_fixture)

    result = invoke([
        'request',
        '--item-type=PSScene',
        '--bundle=visual',
        '--name=test',
        '20220325_131639_20_2402',
        f'--clip={json.dumps(geom)}',
    ])
    assert result.exit_code == 0

    order_request = {
        "name": "test",
        "products": [{
            "item_ids": ["20220325_131639_20_2402"],
            "item_type": "PSScene",
            "product_bundle": "visual",
        }],
        "tools": [{
            'clip': {
                'aoi': geom
            }
        }],
        "metadata": stac_json
    }
    assert order_request == json.loads(result.output)


def test_cli_orders_request_clip_multipolygon(mock_bundles,
                                              multipolygon_geom_geojson,
                                              invoke,
                                              geom_geojson,
                                              stac_json):
    result = invoke([
        'request',
        '--item-type=PSScene',
        '--bundle=visual',
        '--name=test',
        '20220325_131639_20_2402',
        f'--clip={json.dumps(multipolygon_geom_geojson)}',
    ])
    assert result.exit_code == 0

    order_request = {
        "name": "test",
        "products": [{
            "item_ids": ["20220325_131639_20_2402"],
            "item_type": "PSScene",
            "product_bundle": "visual",
        }],
        "tools": [{
            'clip': {
                'aoi': multipolygon_geom_geojson
            }
        }],
        "metadata": stac_json
    }
    assert order_request == json.loads(result.output)


def test_cli_orders_request_clip_invalid_geometry(mock_bundles,
                                                  invoke,
                                                  point_geom_geojson):
    result = invoke([
        'request',
        '--item-type=PSScene',
        '--bundle=visual',
        '--name=test',
        '20220325_131639_20_2402',
        f'--clip={json.dumps(point_geom_geojson)}'
    ])
    assert result.exit_code == 2


def test_cli_orders_request_both_clip_and_tools(mock_bundles,
                                                invoke,
                                                geom_geojson):
    # interestingly, it is important that both clip and tools
    # option values are valid json
    result = invoke([
        'request',
        '--item-type=PSScene',
        '--bundle=visual',
        '--name=test',
        '20220325_131639_20_2402',
        f'--clip={json.dumps(geom_geojson)}',
        f'--tools={json.dumps(geom_geojson)}'
    ])

    assert result.exit_code == 2
    assert "Specify only one of '--clip' or '--tools'" in result.output


def test_cli_orders_request_cloudconfig(mock_bundles, invoke, stac_json):
    config_json = {
        'amazon_s3': {
            'aws_access_key_id': 'aws_access_key_id',
            'aws_secret_access_key': 'aws_secret_access_key',
            'bucket': 'bucket',
            'aws_region': 'aws_region'
        },
        'archive_type': 'zip'
    }

    result = invoke([
        'request',
        '--item-type=PSScene',
        '--bundle=visual',
        '--name=test',
        '20220325_131639_20_2402',
        f'--cloudconfig={json.dumps(config_json)}',
    ])
    assert result.exit_code == 0

    order_request = {
        "name": "test",
        "products": [{
            "item_ids": ["20220325_131639_20_2402"],
            "item_type": "PSScene",
            "product_bundle": "visual",
        }],
        "delivery": config_json,
        "metadata": stac_json
    }
    assert order_request == json.loads(result.output)


def test_cli_orders_request_email(mock_bundles, invoke, stac_json):
    result = invoke([
        'request',
        '--item-type=PSScene',
        '--bundle=visual',
        '--name=test',
        '20220325_131639_20_2402',
        '--email'
    ])
    assert result.exit_code == 0

    order_request = {
        "name": "test",
        "products": [{
            "item_ids": ["20220325_131639_20_2402"],
            "item_type": "PSScene",
            "product_bundle": "visual",
        }],
        "notifications": {
            "email": True,
        },
        "metadata": stac_json
    }
    assert order_request == json.loads(result.output)


@respx.mock
def test_cli_orders_request_tools(mock_bundles,
                                  invoke,
                                  geom_geojson,
                                  stac_json):

    tools_json = [{'clip': {'aoi': geom_geojson}}, {'composite': {}}]

    result = invoke([
        'request',
        '--item-type=PSScene',
        '--bundle=visual',
        '--name=test',
        '20220325_131639_20_2402',
        f'--tools={json.dumps(tools_json)}'
    ])

    order_request = {
        "name": "test",
        "products": [{
            "item_ids": ["20220325_131639_20_2402"],
            "item_type": "PSScene",
            "product_bundle": "visual",
        }],
        "tools": tools_json,
        "metadata": stac_json
    }
    assert order_request == json.loads(result.output)


@respx.mock
def test_cli_orders_request_no_stac(mock_bundles, invoke):

    result = invoke([
        'request',
        '--item-type=PSScene',
        '--bundle=visual',
        '--name=test',
        '20220325_131639_20_2402',
        '--no-stac'
    ])

    order_request = {
        "name": "test",
        "products": [{
            "item_ids": ["20220325_131639_20_2402"],
            "item_type": "PSScene",
            "product_bundle": "visual",
        }]
    }
    assert order_request == json.loads(result.output)


@respx.mock
def test_cli_orders_request_hosting_sentinel_hub(mock_bundles,
                                                 invoke,
                                                 stac_json):

    result = invoke([
        'request',
        '--item-type=PSScene',
        '--bundle=visual',
        '--name=test',
        '20220325_131639_20_2402',
        '--hosting=sentinel_hub',
    ])

    order_request = {
        "name": "test",
        "products": [{
            "item_ids": ["20220325_131639_20_2402"],
            "item_type": "PSScene",
            "product_bundle": "visual",
        }],
        "metadata": stac_json,
        "hosting": {
            "sentinel_hub": {}
        }
    }
    assert order_request == json.loads(result.output)


@respx.mock
def test_cli_orders_request_hosting_sentinel_hub_collection_id(
        mock_bundles, invoke, stac_json):

    result = invoke([
        'request',
        '--item-type=PSScene',
        '--bundle=visual',
        '--name=test',
        '20220325_131639_20_2402',
        '--hosting=sentinel_hub',
        '--collection-id=1234'
    ])

    order_request = {
        "name": "test",
        "products": [{
            "item_ids": ["20220325_131639_20_2402"],
            "item_type": "PSScene",
            "product_bundle": "visual",
        }],
        "metadata": stac_json,
        "hosting": {
            "sentinel_hub": {
                "collection_id": "1234"
            }
        }
    }
    assert order_request == json.loads(result.output)


def test_cli_orders_request_hosting_sentinel_hub_create_configuration(
        invoke, stac_json):

    result = invoke([
        'request',
        '--item-type=PSScene',
        '--bundle=visual',
        '--name=test',
        '20220325_131639_20_2402',
        '--hosting=sentinel_hub',
        '--create-configuration'
    ])

    order_request = {
        "name": "test",
        "products": [{
            "item_ids": ["20220325_131639_20_2402"],
            "item_type": "PSScene",
            "product_bundle": "visual",
        }],
        "metadata": stac_json,
        "hosting": {
            "sentinel_hub": {
                "create_configuration": True
            }
        }
    }
    assert order_request == json.loads(result.output)


@respx.mock
def test_cli_orders_request_hosting_sentinel_hub_collection_configuration(
        invoke, stac_json):
    # Note, this behavior will be rejected by the API, but it is valid in building a request
    result = invoke([
        'request',
        '--item-type=PSScene',
        '--bundle=visual',
        '--name=test',
        '20220325_131639_20_2402',
        '--hosting=sentinel_hub',
        '--collection-id=1234',
        '--create-configuration'
    ])

    order_request = {
        "name": "test",
        "products": [{
            "item_ids": ["20220325_131639_20_2402"],
            "item_type": "PSScene",
            "product_bundle": "visual",
        }],
        "metadata": stac_json,
        "hosting": {
            "sentinel_hub": {
                "collection_id": "1234", "create_configuration": True
            }
        }
    }
    assert order_request == json.loads(result.output)


@respx.mock
def test_item_types(invoke, mock_bundles):

    result = invoke(['item-types'])

    expected_item_types = ["SkySatScene", "SkySatCollect", "PSScene"]
    for item_type in expected_item_types:
        assert item_type in result.output
    assert result.exit_code == 0


@respx.mock
def test_bundle_names(invoke, mock_bundles):

    result = invoke(['bundles'])

    expected_item_types = [
        "analytic_udm2", "analytic_sr", "analytic", "visual"
    ]
    for item_type in expected_item_types:
        assert item_type in result.output
    assert result.exit_code == 0



# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Example of downloading multiple assets in parallel

This is an example of getting two assets, activating them, waiting for them to
become active, downloading them, then validating the checksums of downloaded
files.

[Planet Explorer](https://www.planet.com/explorer/) was used to define
the AOIs and get the image ids.
"""
import asyncio

from planet import Session

river_item_id = '20221003_002705_38_2461'
river_item_type = 'PSScene'
river_asset_type = 'ortho_analytic_4b'

wildfire_item_id = '20221019_183717_11_2475'
wildfire_item_type = 'PSScene'
wildfire_asset_type = 'basic_analytic_4b'


async def download_and_validate(client, item_id, item_type_id, asset_type_id):
    """Activate, download, and validate an asset as a single task."""
    # Get asset description
    asset = await client.get_asset(item_type_id, item_id, asset_type_id)

    # Activate asset
    await client.activate_asset(asset)

    # Wait for asset to become active
    asset = await client.wait_asset(asset, callback=print)

    # Download asset
    path = await client.download_asset(asset)

    # Validate download file
    client.validate_checksum(asset, path)


async def main():
    """Download and validate assets in parallel."""
    async with Session() as sess:
        client = sess.client('data')
        await asyncio.gather(
            download_and_validate(client,
                                  river_item_id,
                                  river_item_type,
                                  river_asset_type),
            download_and_validate(client,
                                  wildfire_item_id,
                                  wildfire_item_type,
                                  wildfire_asset_type))


if __name__ == '__main__':
    asyncio.run(main())



# Copyright 2022 Planet Labs, PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""CLI Parameter types"""
from datetime import datetime
import json
from typing import List

import click

from planet import exceptions, io


class CommaSeparatedString(click.types.StringParamType):
    """A list of strings that is extracted from a comma-separated string."""

    def convert(self, value, param, ctx) -> List[str]:
        if isinstance(value, list):
            convlist = value
        else:
            convstr = super().convert(value, param, ctx)

            if convstr == '':
                self.fail('Entry cannot be an empty string.')

            convlist = [part.strip() for part in convstr.split(",")]

            for v in convlist:
                if not v:
                    self.fail(f'Empty entry encountered in "{value}".')

        return convlist


class CommaSeparatedFloat(click.types.StringParamType):
    """A list of floats that is extracted from a comma-separated string."""
    name = 'VALUE'

    def convert(self, value, param, ctx) -> List[float]:
        values = CommaSeparatedString().convert(value, param, ctx)

        try:
            ret = [float(v) for v in values]
        except ValueError:
            self.fail(f'Cound not convert all entries in "{value}" to float.')

        return ret


class JSON(click.ParamType):
    """JSON specified as a string, json file filename, or stdin."""
    name = 'JSON'

    def convert(self, value, param, ctx) -> dict:
        if isinstance(value, dict):
            convdict = value
        else:
            # read from raw json
            # skip this if value is a Path object or something else
            if isinstance(value, str) and (value.startswith('{')
                                           or value.startswith('[')):
                try:
                    convdict = json.loads(value)
                except json.decoder.JSONDecodeError:
                    self.fail(
                        'JSON string is not valid JSON. Make sure the entire '
                        'JSON string is single-quoted and string entries are '
                        'double-quoted.')

            # read from stdin or file
            else:
                try:
                    with click.open_file(value) as f:
                        convdict = json.load(f)
                except FileNotFoundError:
                    self.fail('File not found.')
                except json.decoder.JSONDecodeError:
                    self.fail('JSON file does not contain valid JSON.')

        if convdict == {}:
            self.fail('JSON cannot be empty.')

        return convdict


class Geometry(click.ParamType):
    name = 'geom'

    def __init__(self):
        self.types = [JSON(), CommaSeparatedString()]

    def convert(self, value, param, ctx):
        for type in self.types:
            try:
                return type.convert(value, param, ctx)
            except click.BadParameter:
                continue


class Field(click.ParamType):
    """Clarify that this entry is for a field"""
    name = 'field'


class Comparison(click.ParamType):
    name = 'comp'
    valid = ['lt', 'lte', 'gt', 'gte']

    def convert(self, value, param, ctx) -> str:
        if value not in self.valid:
            self.fail(f'COMP ({value}) must be one of {",".join(self.valid)}',
                      param,
                      ctx)
        return value


class GTComparison(Comparison):
    """Only support gt or gte comparison"""
    valid = ['gt', 'gte']


class DateTime(click.ParamType):
    name = 'datetime'

    def convert(self, value, param, ctx) -> datetime:
        if not isinstance(value, datetime):
            try:
                value = io.str_to_datetime(value)
            except exceptions.PlanetError as e:
                self.fail(str(e))

        return value



# Copyright 2022 Planet Labs, PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""CLI Parameter validation"""
from typing import Optional
from planet import geojson


def check_geom(ctx, param, geometry) -> Optional[dict]:
    """Validates geometry as GeoJSON or feature ref(s)."""
    if isinstance(geometry, dict):
        return geojson.as_geom_or_ref(geometry)
    geoms = {}
    if geometry:
        for geom in geometry:
            geoms.update(geojson.as_geom_or_ref(geom))
    return geoms if geoms else None



# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Orders API CLI"""
from contextlib import asynccontextmanager
import logging
from pathlib import Path
from typing import List, Optional

import click

import planet
from planet import OrdersClient  # allow mocking
from . import types
from .cmds import coro, translate_exceptions
from ..order_request import sentinel_hub
from .io import echo_json
from .options import limit, pretty
from .session import CliSession
from ..specs import (FetchBundlesSpecError,
                     get_bundle_names,
                     get_item_types,
                     SpecificationException,
                     validate_bundle,
                     validate_data_item_type)

LOGGER = logging.getLogger(__name__)


def check_item_type(ctx, param, item_type):
    """Validates the item type provided by comparing it to all supported
    item types."""
    try:
        validate_data_item_type(item_type)
        return item_type
    except SpecificationException as e:
        raise click.BadParameter(str(e))
    except FetchBundlesSpecError as e:
        raise click.ClickException(str(e))


def check_bundle(ctx, param, bundle) -> Optional[List[dict]]:
    """Validates the bundle provided by comparing it to all supported
    bundles."""

    # Get the value of item_type from the context
    item_type = ctx.params.get('item_type')
    if not item_type:
        raise click.BadParameter("Item type is required to validate a bundle.")

    if bundle is None:
        return bundle

    bundles = bundle.split(",")
    for b in bundles:
        try:
            validate_bundle(item_type, b)
        except SpecificationException as e:
            raise click.BadParameter(str(e))
        except FetchBundlesSpecError as e:
            raise click.ClickException(str(e))

    return bundle


@asynccontextmanager
async def orders_client(ctx):
    base_url = ctx.obj['BASE_URL']
    async with CliSession() as sess:
        cl = OrdersClient(sess, base_url=base_url)
        yield cl


@click.group()  # type: ignore
@click.pass_context
@click.option('-u',
              '--base-url',
              default=None,
              help='Assign custom base Orders API URL.')
def orders(ctx, base_url):
    """Commands for interacting with the Orders API"""
    ctx.obj['BASE_URL'] = base_url


@orders.command()  # type: ignore
@click.pass_context
@translate_exceptions
@coro
@click.option('--state',
              help='Filter by state.',
              type=click.Choice(planet.clients.orders.ORDER_STATE_SEQUENCE,
                                case_sensitive=False))
@click.option(
    '--source-type',
    default="all",
    help="""Filter by source type. See documentation for all available types.
    Default is all.""")
@click.option('--name', help="filter by name")
@click.option(
    '--name-contains',
    help="only return orders with a name that contains the provided string.")
@click.option(
    '--created-on',
    help="Filter by creation time or interval. See documentation for examples."
)
@click.option(
    '--last-modified',
    help="Filter by creation time or interval. See documentation for examples."
)
@click.option(
    '--hosting',
    type=click.BOOL,
    help="Filter by presence of hosting block (e.g. SentinelHub hosting).")
@click.option(
    '--sort-by',
    help="""fields to sort orders by. Multiple fields can be specified,
    separated by commas. The sort direction can be specified by appending
    ' ASC' or ' DESC' to the field name. The default sort direction is
    ascending. When multiple fields are specified, the sort order is applied
    in the order the fields are listed.

    Supported fields: [name, created_on, state, last_modified].

    Example: 'name ASC,created_on DESC'""")
@limit
@pretty
async def list(ctx,
               state,
               source_type,
               name,
               name_contains,
               created_on,
               last_modified,
               hosting,
               sort_by,
               limit,
               pretty):
    """List orders

    This command prints a sequence of the returned order descriptions,
    optionally pretty-printed.

    By default, order descriptions are sorted by creation date with the last created order
    returned first.
    """
    async with orders_client(ctx) as cl:
        async for o in cl.list_orders(state=state,
                                      source_type=source_type,
                                      name=name,
                                      name__contains=name_contains,
                                      created_on=created_on,
                                      last_modified=last_modified,
                                      hosting=hosting,
                                      sort_by=sort_by,
                                      limit=limit):
            echo_json(o, pretty)


@orders.command()  # type: ignore
@click.pass_context
@translate_exceptions
@coro
@click.argument('order_id', type=click.UUID)
@pretty
async def get(ctx, order_id, pretty):
    """Get order

    This command outputs the order description, optionally pretty-printed.
    """
    async with orders_client(ctx) as cl:
        order = await cl.get_order(str(order_id))

    echo_json(order, pretty)


@orders.command()  # type: ignore
@click.pass_context
@translate_exceptions
@coro
@click.argument('order_id', type=click.UUID)
async def cancel(ctx, order_id):
    """Cancel order by order ID.

    This command cancels a queued order and outputs the cancelled order
    details.
    """
    async with orders_client(ctx) as cl:
        json_resp = await cl.cancel_order(str(order_id))

    click.echo(json_resp)


@orders.command()  # type: ignore
@click.pass_context
@translate_exceptions
@coro
@click.argument('order_id', type=click.UUID)
@click.option('--delay',
              type=int,
              default=5,
              help='Time (in seconds) between polls.')
@click.option('--max-attempts',
              type=int,
              default=200,
              show_default=True,
              help='Maximum number of polls. Set to zero for no limit.')
@click.option('--state',
              help='State prior to a final state that will end polling.',
              type=click.Choice(planet.clients.orders.ORDER_STATE_SEQUENCE,
                                case_sensitive=False))
async def wait(ctx, order_id, delay, max_attempts, state):
    """Wait until order reaches desired state.

    Reports the state of the order on the last poll.

    This function polls the Orders API to determine the order state, with
    the specified delay between each polling attempt, until the
    order reaches a final state or earlier state, if specified.
    If the maximum number of attempts is reached before polling is
    complete, an exception is raised. Setting --max-attempts to zero will
    result in no limit on the number of attempts.

    Setting --delay to zero results in no delay between polling attempts.
    This will likely result in throttling by the Orders API, which has
    a rate limit of 10 requests per second. If many orders are being
    polled asynchronously, consider increasing the delay to avoid
    throttling.

    By default, polling completes when the order reaches a final state.
    If --state is specified, polling will complete when the specified earlier
    state is reached or passed.
    """
    quiet = ctx.obj['QUIET']
    async with orders_client(ctx) as cl:
        with planet.reporting.StateBar(order_id=order_id,
                                       disable=quiet) as bar:
            state = await cl.wait(str(order_id),
                                  state=state,
                                  delay=delay,
                                  max_attempts=max_attempts,
                                  callback=bar.update_state)
    click.echo(state)


@orders.command()  # type: ignore
@click.pass_context
@translate_exceptions
@coro
@click.argument('order_id', type=click.UUID)
@click.option('--checksum',
              default=None,
              type=click.Choice(['MD5', 'SHA256'], case_sensitive=False),
              help=('Verify that checksums match.'))
@click.option('--directory',
              default='.',
              help=('Base directory for file download.'),
              type=click.Path(exists=True,
                              resolve_path=True,
                              writable=True,
                              file_okay=False))
@click.option('--overwrite',
              is_flag=True,
              default=False,
              help=('Overwrite files if they already exist.'))
async def download(ctx, order_id, overwrite, directory, checksum):
    """Download order by order ID.

    If --checksum is provided, the associated checksums given in the manifest
    are compared against the downloaded files to verify that they match.

    If --checksum is provided, files are already downloaded, and --overwrite is
    not specified, this will simply validate the checksums of the files against
    the manifest.
    """
    quiet = ctx.obj['QUIET']
    async with orders_client(ctx) as cl:
        await cl.download_order(
            str(order_id),
            directory=Path(directory),
            overwrite=overwrite,
            progress_bar=not quiet,
        )
        if checksum:
            cl.validate_checksum(Path(directory, str(order_id)), checksum)


@orders.command()  # type: ignore
@click.pass_context
@translate_exceptions
@coro
@click.argument("request", type=types.JSON())
@click.option(
    "--hosting",
    type=click.Choice([
        "sentinel_hub",
    ]),
    default=None,
    help='Hosting type. Currently, only "sentinel_hub" is supported.',
)
@click.option("--collection-id",
              default=None,
              help='Collection ID for Sentinel Hub hosting. '
              'If omitted, a new collection will be created.')
@click.option(
    '--create-configuration',
    is_flag=True,
    default=False,
    help='Automatically create a layer configuration for your collection. '
    'If omitted, no configuration will be created.')
@pretty
async def create(ctx, request, pretty, **kwargs):
    """Create an order.

    This command outputs the created order description, optionally
    pretty-printed.

    REQUEST is the full description of the order to be created. It must be JSON
    and can be specified a json string, filename, or '-' for stdin.

    Other flag options are hosting, collection_id, and create_configuration.
    The hosting flag specifies the hosting type, the collection_id flag specifies the
    collection ID for Sentinel Hub, and the create_configuration flag specifies
    whether or not to create a layer configuration for your collection. If the
    collection_id is omitted, a new collection will be created. If the
    create_configuration flag is omitted, no configuration will be created. The
    collection_id flag and create_configuration flag cannot be used together.
    """

    hosting = kwargs.get('hosting')
    collection_id = kwargs.get('collection_id')
    create_configuration = bool(kwargs.get('create_configuration', False))

    if hosting == "sentinel_hub":
        request["hosting"] = sentinel_hub(collection_id, create_configuration)

    async with orders_client(ctx) as cl:
        order = await cl.create_order(request)

    echo_json(order, pretty)


@orders.command()  # type: ignore
@click.pass_context
@translate_exceptions
@coro
@click.argument('ids', metavar='IDS', type=types.CommaSeparatedString())
@click.option('--item-type',
              required=True,
              help='Item type for requested item ids.',
              type=str,
              callback=check_item_type)
@click.option('--bundle',
              required=True,
              help='Bundle type for the item.',
              type=str,
              callback=check_bundle)
@click.option(
    '--fallback-bundle',
    required=False,
    help="""Fallback bundle type(s) for the item if bundle is not available.
    Multiple fallback bundles can be specified as a comma-separated string.""",
    type=str,
    callback=check_bundle)
@click.option('--name',
              required=True,
              help='Order name. Does not need to be unique.',
              type=click.STRING)
@click.option('--clip',
              type=types.JSON(),
              help="""Clip feature Polygon or Multipolygon GeoJSON. Can be a
              json string, filename, or '-' for stdin.""")
@click.option(
    '--tools',
    type=types.JSON(),
    help="""Toolchain JSON. Can be a json string, filename, or '-' for
    stdin.""")
@click.option('--email',
              default=False,
              is_flag=True,
              help='Send email notification when order is complete.')
@click.option('--archive-type',
              type=click.Choice(['zip']),
              help="Optionally zip archive each item bundle.")
@click.option('--archive-filename',
              default='{{name}}-{{order_id}}.zip',
              show_default=True,
              help="Templated filename for archived bundles or orders.")
@click.option('--single-archive',
              is_flag=True,
              default=False,
              show_default=True,
              help="Optionally zip archive all item bundles together.")
@click.option(
    '--delivery',
    '--cloudconfig',
    type=types.JSON(),
    help=("Delivery configuration, which may include credentials for a cloud "
          "storage provider, to enable cloud delivery of data, and/or "
          "parameters for bundling deliveries as zip archives. Can be a JSON "
          "string, a filename, or '-' for stdin. The --cloudconfig option is "
          "an alias for this use case."))
@click.option(
    '--stac/--no-stac',
    default=True,
    is_flag=True,
    help="""Include or exclude metadata in SpatioTemporal Asset Catalog (STAC)
    format. Not specifying either defaults to including it (--stac), except
    for orders with google_earth_engine delivery""")
@click.option('--hosting',
              type=click.Choice(['sentinel_hub']),
              help='Hosting for data delivery. '
              'Currently, only "sentinel_hub" is supported.')
@click.option('--collection-id',
              help='Collection ID for Sentinel Hub hosting. '
              'If omitted, a new collection will be created.')
@click.option(
    '--create-configuration',
    is_flag=True,
    help='Automatically create a layer configuration for your collection. '
    'If omitted, no configuration will be created.')
@pretty
async def request(ctx,
                  item_type,
                  bundle,
                  fallback_bundle,
                  name,
                  ids,
                  clip,
                  tools,
                  email,
                  archive_type,
                  archive_filename,
                  single_archive,
                  delivery,
                  stac,
                  hosting,
                  collection_id,
                  create_configuration,
                  pretty):
    """Generate an order request.

    This command provides support for building an order description used
    in creating an order. It outputs the order request, optionally
    pretty-printed.

    IDs is one or more comma-separated item IDs.
    """
    try:
        product = planet.order_request.product(ids,
                                               bundle,
                                               item_type,
                                               fallback_bundle)
    except planet.specs.SpecificationException as e:
        raise click.BadParameter(e)

    if email:
        notifications = planet.order_request.notifications(email=email)
    else:
        notifications = None

    if clip and tools:
        raise click.BadParameter("Specify only one of '--clip' or '--tools'")
    elif clip:
        try:
            tools = [planet.order_request.clip_tool(clip)]
        except planet.exceptions.ClientError as e:
            raise click.BadParameter(e)

    delivery = planet.order_request.delivery(archive_type=archive_type,
                                             archive_filename=archive_filename,
                                             single_archive=single_archive,
                                             cloud_config=delivery)

    if stac and "google_earth_engine" not in delivery:
        stac_json = {'stac': {}}
    else:
        stac_json = {}

    request = planet.order_request.build_request(
        name,
        products=[product],
        delivery=delivery,
        notifications=notifications,
        tools=tools,
        stac=stac_json,
        hosting=hosting,
        collection_id=collection_id,
        create_configuration=create_configuration)

    echo_json(request, pretty)


@orders.command()  # type: ignore
@click.pass_context
@translate_exceptions
def item_types(ctx):
    """Show valid item types for ordering."""
    click.echo("Valid item types:")
    for it in get_item_types():
        click.echo(f"- {it}")


@orders.command()  # type: ignore
@click.pass_context
@translate_exceptions
def bundles(ctx):
    """Show valid bundle names for ordering."""
    click.echo("Valid bundles:")
    for it in get_bundle_names():
        click.echo(f"- {it}")



"""Planet Subscriptions API Python client."""

import logging
from typing import Any, AsyncIterator, Awaitable, Dict, Optional, Sequence, TypeVar

from typing_extensions import Literal

from planet.exceptions import APIError, ClientError
from planet.http import Session
from planet.models import Paged
from ..constants import PLANET_BASE_URL

BASE_URL = f'{PLANET_BASE_URL}/subscriptions/v1/'

LOGGER = logging.getLogger()

T = TypeVar("T")


class SubscriptionsClient:
    """A Planet Subscriptions Service API 1.0.0 client.

    The methods of this class forward request parameters to the
    operations described in the Planet Subscriptions Service API 1.0.0
    (https://api.planet.com/subscriptions/v1/spec) using HTTP 1.1.

    The methods generally return or yield Python dicts with the same
    structure as the JSON messages used by the service API. Many of the
    exceptions raised by this class are categorized by HTTP client
    (4xx) or server (5xx) errors. This client's level of abstraction is
    low.

    High-level asynchronous access to Planet's subscriptions API:

    Example:
        ```python
        >>> import asyncio
        >>> from planet import Session
        >>>
        >>> async def main():
        ...     async with Session() as sess:
        ...         cl = sess.client('subscriptions')
        ...         # use client here
        ...
        >>> asyncio.run(main())
        ```
    """

    def __init__(self,
                 session: Session,
                 base_url: Optional[str] = None) -> None:
        """
        Parameters:
            session: Open session connected to server.
            base_url: The base URL to use. Defaults to production subscriptions
                API base url.
        """
        self._session = session

        self._base_url = base_url or BASE_URL
        if self._base_url.endswith('/'):
            self._base_url = self._base_url[:-1]

    def _call_sync(self, f: Awaitable[T]) -> T:
        """block on an async function call, using the call_sync method of the session"""
        return self._session._call_sync(f)

    async def list_subscriptions(self,
                                 status: Optional[Sequence[str]] = None,
                                 limit: int = 100,
                                 created: Optional[str] = None,
                                 end_time: Optional[str] = None,
                                 hosting: Optional[bool] = None,
                                 name__contains: Optional[str] = None,
                                 name: Optional[str] = None,
                                 source_type: Optional[str] = None,
                                 start_time: Optional[str] = None,
                                 sort_by: Optional[str] = None,
                                 updated: Optional[str] = None,
                                 page_size: int = 500) -> AsyncIterator[dict]:
        """Iterate over list of account subscriptions with optional filtering.

        Note:
            The name of this method is based on the API's method name.
            This method provides iteration over subscriptions, it does
            not return a list.

        Args:
            created (str): filter by created time or interval.
            end_time (str): filter by end time or interval.
            hosting (bool): only return subscriptions that contain a
                hosting block (e.g. SentinelHub hosting).
            name__contains (str): only return subscriptions with a
                name that contains the given string.
            name (str): filter by name.
            source_type (str): filter by source type.
            start_time (str): filter by start time or interval.
            status (Set[str]): include subscriptions with a status in this set.
            sort_by (str): fields to sort subscriptions by. Multiple
                fields can be specified, separated by commas. The sort
                direction can be specified by appending ' ASC' or '
                DESC' to the field name. The default sort direction is
                ascending. When multiple fields are specified, the sort
                order is applied in the order the fields are listed.

                Supported fields: name, created, updated, start_time, end_time

                Examples:
                 * "name"
                 * "name DESC"
                 * "name,end_time DESC,start_time"
            updated (str): filter by updated time or interval.
            limit (int): limit the number of subscriptions in the
                results. When set to 0, no maximum is applied.
            page_size (int): number of subscriptions to return per page.
            TODO: user_id

        Datetime args (created, end_time, start_time, updated) can either be a
        date-time or an interval, open or closed. Date and time expressions adhere
        to RFC 3339. Open intervals are expressed using double-dots.

        Examples:
            * A date-time: "2018-02-12T23:20:50Z"
            * A closed interval: "2018-02-12T00:00:00Z/2018-03-18T12:31:12Z"
            * Open intervals: "2018-02-12T00:00:00Z/.." or "../2018-03-18T12:31:12Z"

        Yields:
            dict: a description of a subscription.

        Raises:
            APIError: on an API server error.
            ClientError: on a client error.
        """

        class _SubscriptionsPager(Paged):
            """Navigates pages of messages about subscriptions."""
            ITEMS_KEY = 'subscriptions'

        params: Dict[str, Any] = {}
        if created is not None:
            params['created'] = created
        if end_time is not None:
            params['end_time'] = end_time
        if hosting is not None:
            params['hosting'] = hosting
        if name__contains is not None:
            params['name__contains'] = name__contains
        if name is not None:
            params['name'] = name
        if source_type is not None:
            params['source_type'] = source_type
        if start_time is not None:
            params['start_time'] = start_time
        if status is not None:
            params['status'] = [val for val in status]
        if sort_by is not None:
            params['sort_by'] = sort_by
        if updated is not None:
            params['updated'] = updated

        params['page_size'] = page_size

        try:
            response = await self._session.request(method='GET',
                                                   url=self._base_url,
                                                   params=params)
            async for sub in _SubscriptionsPager(response,
                                                 self._session.request,
                                                 limit=limit):
                yield sub
        # Forward APIError. We don't strictly need this clause, but it
        # makes our intent clear.
        except APIError:
            raise
        except ClientError:  # pragma: no cover
            raise

    async def create_subscription(self, request: dict) -> dict:
        """Create a Subscription.

        Args:
            request (dict): description of a subscription.

        Returns:
            dict: description of created subscription.

        Raises:
            APIError: on an API server error.
            ClientError: on a client error.
        """

        try:
            resp = await self._session.request(method='POST',
                                               url=self._base_url,
                                               json=request)
        # Forward APIError. We don't strictly need this clause, but it
        # makes our intent clear.
        except APIError:
            raise
        except ClientError:  # pragma: no cover
            raise
        else:
            sub = resp.json()
            return sub

    async def cancel_subscription(self, subscription_id: str) -> None:
        """Cancel a Subscription.

        Args:
            subscription_id (str): id of subscription to cancel.

        Returns:
            None

        Raises:
            APIError: on an API server error.
            ClientError: on a client error.
        """
        url = f'{self._base_url}/{subscription_id}/cancel'

        try:
            _ = await self._session.request(method='POST', url=url)
        # Forward APIError. We don't strictly need this clause, but it
        # makes our intent clear.
        except APIError:
            raise
        except ClientError:  # pragma: no cover
            raise

    async def update_subscription(self, subscription_id: str,
                                  request: dict) -> dict:
        """Update (edit) a Subscription via PUT.

        Args
            subscription_id (str): id of the subscription to update.
            request (dict): subscription content for update, full
                payload is required.

        Returns:
            dict: description of the updated subscription.

        Raises:
            APIError: on an API server error.
            ClientError: on a client error.
        """
        url = f'{self._base_url}/{subscription_id}'

        try:
            resp = await self._session.request(method='PUT',
                                               url=url,
                                               json=request)
        # Forward APIError. We don't strictly need this clause, but it
        # makes our intent clear.
        except APIError:
            raise
        except ClientError:  # pragma: no cover
            raise
        else:
            sub = resp.json()
            return sub

    async def patch_subscription(self, subscription_id: str,
                                 request: dict) -> dict:
        """Update (edit) a Subscription via PATCH.

        Args
            subscription_id (str): id of the subscription to update.
            request (dict): subscription content for update, only
                attributes to update are required.

        Returns:
            dict: description of the updated subscription.

        Raises:
            APIError: on an API server error.
            ClientError: on a client error.
        """
        url = f'{self._base_url}/{subscription_id}'

        try:
            resp = await self._session.request(method='PATCH',
                                               url=url,
                                               json=request)
        # Forward APIError. We don't strictly need this clause, but it
        # makes our intent clear.
        except APIError:
            raise
        except ClientError:  # pragma: no cover
            raise
        else:
            sub = resp.json()
            return sub

    async def get_subscription(self, subscription_id: str) -> dict:
        """Get a description of a Subscription.

        Args:
            subscription_id (str): id of a subscription.

        Returns:
            dict: description of the subscription.

        Raises:
            APIError: on an API server error.
            ClientError: on a client error.
        """
        url = f'{self._base_url}/{subscription_id}'

        try:
            resp = await self._session.request(method='GET', url=url)
        # Forward APIError. We don't strictly need this clause, but it
        # makes our intent clear.
        except APIError:
            raise
        except ClientError:  # pragma: no cover
            raise
        else:
            sub = resp.json()
            return sub

    async def get_results(self,
                          subscription_id: str,
                          status: Optional[Sequence[Literal[
                              "created",
                              "queued",
                              "processing",
                              "failed",
                              "success"]]] = None,
                          limit: int = 100) -> AsyncIterator[dict]:
        """Iterate over results of a Subscription.

        Notes:
            The name of this method is based on the API's method name. This
            method provides iteration over results, it does not get a
            single result description or return a list of descriptions.

        Parameters:
            subscription_id (str): id of a subscription.
            status (Set[str]): pass result with status in this set,
                filter out results with status not in this set.
            limit (int): limit the number of subscriptions in the
                results. When set to 0, no maximum is applied.
            TODO: created, updated, completed, user_id

        Yields:
            dict: description of a subscription results.

        Raises:
            APIError: on an API server error.
            ClientError: on a client error.
        """

        class _ResultsPager(Paged):
            """Navigates pages of messages about subscription results."""
            ITEMS_KEY = 'results'

        params = {'status': [val for val in status or {}]}
        url = f'{self._base_url}/{subscription_id}/results'

        try:
            resp = await self._session.request(method='GET',
                                               url=url,
                                               params=params)
            async for sub in _ResultsPager(resp,
                                           self._session.request,
                                           limit=limit):
                yield sub
        # Forward APIError. We don't strictly need this clause, but it
        # makes our intent clear.
        except APIError:
            raise
        except ClientError:  # pragma: no cover
            raise

    async def get_results_csv(
        self,
        subscription_id: str,
        status: Optional[Sequence[Literal["created",
                                          "queued",
                                          "processing",
                                          "failed",
                                          "success"]]] = None
    ) -> AsyncIterator[str]:
        """Iterate over rows of results CSV for a Subscription.

        Parameters:
            subscription_id (str): id of a subscription.
            status (Set[str]): pass result with status in this set,
                filter out results with status not in this set.
            TODO: created, updated, completed, user_id

        Yields:
            str: a row from a CSV file.

        Raises:
            APIError: on an API server error.
            ClientError: on a client error.
        """
        url = f'{self._base_url}/{subscription_id}/results'
        params = {'status': [val for val in status or {}], 'format': 'csv'}

        # Note: retries are not implemented yet. This project has
        # retry logic for HTTP requests, but does not handle errors
        # during streaming. We may want to consider a retry decorator
        # for this entire method a la stamina:
        # https://github.com/hynek/stamina.
        async with self._session.stream('GET', url, params=params) as response:
            async for line in response.aiter_lines():
                yield line

    async def get_summary(self) -> dict:
        """Summarize the status of all subscriptions via GET.

        Returns:
            dict: subscription totals by status.

        Raises:
            APIError: on an API server error.
            ClientError: on a client error.
        """
        url = f'{self._base_url}/summary'

        try:
            resp = await self._session.request(method='GET', url=url)
        # Forward APIError. We don't strictly need this clause, but it
        # makes our intent clear.
        except APIError:
            raise
        except ClientError:  # pragma: no cover
            raise
        else:
            summary = resp.json()
            return summary

    async def get_subscription_summary(
        self,
        subscription_id: str,
    ) -> dict:
        """Summarize the status of results for a single subscription via GET.

        Parameters:
            subscription_id (str): ID of the subscription to summarize.

        Returns:
            dict: result totals for the provided subscription by status.

        Raises:
            APIError: on an API server error.
            ClientError: on a client error.
        """
        url = f'{self._base_url}/{subscription_id}/summary'

        try:
            resp = await self._session.request(method='GET', url=url)
        # Forward APIError. We don't strictly need this clause, but it
        # makes our intent clear.
        except APIError:
            raise
        except ClientError:  # pragma: no cover
            raise
        else:
            summary = resp.json()
            return summary



# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import pytest
import click
import respx
from planet.cli import data
from planet.cli import orders
from planet.cli import subscriptions

LOGGER = logging.getLogger(__name__)


class MockContext:

    def __init__(self):
        self.obj = {}
        self.params = {}


@respx.mock
def test_item_types_success_data(mock_bundles):
    ctx = MockContext()
    result = data.check_item_types(ctx, 'item_types', ["PSScene"])
    assert result == ["PSScene"]


@respx.mock
def test_item_types_fail_data(mock_bundles):
    ctx = MockContext()
    with pytest.raises(click.BadParameter):
        data.check_item_types(ctx, 'item_types', "bad_item_type")


@respx.mock
def test_item_type_success_data(mock_bundles):
    ctx = MockContext()
    item_type = "PSScene"
    result = data.check_item_type(ctx, 'item_type', item_type)
    assert result == item_type


@respx.mock
def test_item_type_fail_data(mock_bundles):
    ctx = MockContext()
    with pytest.raises(click.BadParameter):
        data.check_item_type(ctx, 'item_type', "bad_item_type")


@respx.mock
def test_item_type_too_many_item_types_data(mock_bundles):
    ctx = MockContext()
    with pytest.raises(click.BadParameter):
        data.check_item_types(ctx, 'item_type', "PSScene,SkySatScene")


# Identical tests to above, but for subscriptions CLI
@respx.mock
def test_item_types_success_subscriptions(mock_bundles):
    ctx = MockContext()
    result = subscriptions.check_item_types(ctx, 'item_types', ["PSScene"])
    assert result == ["PSScene"]


@respx.mock
def test_item_types_fail_subscriptions(mock_bundles):
    ctx = MockContext()
    with pytest.raises(click.BadParameter):
        subscriptions.check_item_types(ctx, 'item_types', "bad_item_type")


@respx.mock
def test_item_type_too_many_item_types_subscriptions(mock_bundles):
    ctx = MockContext()
    with pytest.raises(click.BadParameter):
        subscriptions.check_item_types(ctx, 'item_type', "PSScene,SkySatScene")


@respx.mock
def test_item_type_success_orders(mock_bundles):
    ctx = MockContext()
    item_type = "PSScene"
    result = orders.check_item_type(ctx, 'item_type', item_type)
    assert result == item_type


@respx.mock
def test_item_type_fail_orders(mock_bundles):
    ctx = MockContext()
    with pytest.raises(click.BadParameter):
        orders.check_item_type(ctx, 'item_type', "bad_item_type")


@respx.mock
def test_bundle_success_orders(mock_bundles):
    ctx = MockContext()
    ctx.params = {"item_type": "PSScene"}
    bundle = "analytic_udm2"
    result = orders.check_bundle(ctx, 'bundle', bundle)
    assert result == bundle


@respx.mock
def test_bundle_fail_orders(mock_bundles):
    ctx = MockContext()
    ctx.params = {"item_type": "PSScene"}
    with pytest.raises(click.BadParameter):
        orders.check_bundle(ctx, 'bundle', "bad_bundle")
    ctx.params = {"item_type": "bad_item_type"}
    with pytest.raises(click.BadParameter):
        orders.check_bundle(ctx, 'bundle', "analytic_udm2")



# Copyright 2020 Planet Labs, Inc.
# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Functionality for interacting with the orders api"""
import asyncio
import logging
import time
from typing import AsyncIterator, Awaitable, Callable, Dict, List, Optional, Sequence, TypeVar, Union
import uuid
import json
import hashlib

from pathlib import Path
from .. import exceptions
from ..constants import PLANET_BASE_URL
from ..http import Session
from ..models import Paged, StreamingBody

BASE_URL = f'{PLANET_BASE_URL}/compute/ops'
STATS_PATH = '/stats/orders/v2'
ORDERS_PATH = '/orders/v2'
BULK_PATH = '/bulk/orders/v2'

# Order states https://developers.planet.com/apis/orders/states/
# this is in order of state progression except for final states
ORDER_STATE_SEQUENCE = \
    ('queued', 'running', 'failed', 'success', 'partial', 'cancelled')

LOGGER = logging.getLogger(__name__)

T = TypeVar("T")


class Orders(Paged):
    """Asynchronous iterator over Orders from a paged response describing
    orders."""
    ITEMS_KEY = 'orders'


class OrderStates:
    SEQUENCE = ORDER_STATE_SEQUENCE

    @classmethod
    def _get_position(cls, state):
        return cls.SEQUENCE.index(state)

    @classmethod
    def reached(cls, state, test):
        return cls._get_position(test) >= cls._get_position(state)

    @classmethod
    def passed(cls, state, test):
        return cls._get_position(test) > cls._get_position(state)

    @classmethod
    def is_final(cls, test):
        return cls.passed('running', test)


class OrdersClient:
    """High-level asynchronous access to Planet's orders API.

    Example:
        ```python
        >>> import asyncio
        >>> from planet import Session
        >>>
        >>> async def main():
        ...     async with Session() as sess:
        ...         cl = sess.client('orders')
        ...         # use client here
        ...
        >>> asyncio.run(main())

        ```
    """

    def __init__(self, session: Session, base_url: Optional[str] = None):
        """
        Parameters:
            session: Open session connected to server.
            base_url: The base URL to use. Defaults to production orders API
                base url.
        """
        self._session = session

        self._base_url = base_url or BASE_URL
        if self._base_url.endswith('/'):
            self._base_url = self._base_url[:-1]

    def _call_sync(self, f: Awaitable[T]) -> T:
        """block on an async function call, using the call_sync method of the session"""
        return self._session._call_sync(f)

    @staticmethod
    def _check_order_id(oid):
        """Raises planet.exceptions.ClientError if oid is not a valid UUID"""
        try:
            uuid.UUID(hex=oid)
        except (ValueError, AttributeError):
            msg = f'Order id ({oid}) is not a valid UUID hexadecimal string.'
            raise exceptions.ClientError(msg)

    def _orders_url(self):
        return f'{self._base_url}{ORDERS_PATH}'

    def _stats_url(self):
        return f'{self._base_url}{STATS_PATH}'

    async def create_order(self, request: dict) -> dict:
        """Create an order request.

        Example:

        ```python
        >>> import asyncio
        >>> from planet import Session
        >>> from planet.order_request import build_request, product
        >>>
        >>> async def main():
        ...     image_ids = ["20200925_161029_69_2223"]
        ...     request = build_request(
        ...         'test_order',
        ...         [product(image_ids, 'analytic_udm2', 'psscene')]
        ...     )
        ...     async with Session() as sess:
        ...         cl = sess.client('orders')
        ...         order = await cl.create_order(request)
        ...
        >>> asyncio.run(main())
        ```

        Parameters:
            request: order request definition

        Returns:
            JSON description of the created order

        Raises:
            planet.exceptions.APIError: On API error.
        """
        url = self._orders_url()
        response = await self._session.request(method='POST',
                                               url=url,
                                               json=request)
        return response.json()

    async def get_order(self, order_id: str) -> dict:
        """Get order details by Order ID.

        Parameters:
            order_id: The ID of the order

        Returns:
            JSON description of the order

        Raises:
            planet.exceptions.ClientError: If order_id is not a valid UUID.
            planet.exceptions.APIError: On API error.
        """
        self._check_order_id(order_id)
        url = f'{self._orders_url()}/{order_id}'

        response = await self._session.request(method='GET', url=url)
        return response.json()

    async def cancel_order(self, order_id: str) -> dict:
        """Cancel a queued order.

        Parameters:
            order_id: The ID of the order

        Returns:
            Results of the cancel request

        Raises:
            planet.exceptions.ClientError: If order_id is not a valid UUID.
            planet.exceptions.APIError: On API error.
        """
        self._check_order_id(order_id)
        url = f'{self._orders_url()}/{order_id}'

        response = await self._session.request(method='PUT', url=url)
        return response.json()

    async def cancel_orders(self,
                            order_ids: Optional[List[str]] = None) -> dict:
        """Cancel queued orders in bulk.

        Parameters:
            order_ids: The IDs of the orders. If empty or None, all orders in a
                pre-running state will be cancelled.

        Returns:
            Results of the bulk cancel request

        Raises:
            planet.exceptions.ClientError: If an entry in order_ids is not a
                valid UUID.
            planet.exceptions.APIError: On API error.
        """
        url = f'{self._base_url}{BULK_PATH}/cancel'
        cancel_body = {}
        if order_ids:
            for oid in order_ids:
                self._check_order_id(oid)
            cancel_body['order_ids'] = order_ids

        response = await self._session.request(method='POST',
                                               url=url,
                                               json=cancel_body)

        return response.json()

    async def aggregated_order_stats(self) -> dict:
        """Get aggregated counts of active orders.

        Returns:
            Aggregated order counts

        Raises:
            planet.exceptions.APIError: On API error.
        """
        url = self._stats_url()
        response = await self._session.request(method='GET', url=url)
        return response.json()

    async def download_asset(self,
                             location: str,
                             filename: Optional[str] = None,
                             directory: Path = Path('.'),
                             overwrite: bool = False,
                             progress_bar: bool = True) -> Path:
        """Download ordered asset.

        Parameters:
            location: Download location url including download token.
            filename: Custom name to assign to downloaded file.
            directory: Base directory for file download. This directory will be
                created if it does not already exist.
            overwrite: Overwrite any existing files.
            progress_bar: Show progress bar during download.

        Returns:
            Path to downloaded file.

        Raises:
            planet.exceptions.APIError: On API error.
        """
        async with self._session.stream(method='GET', url=location) as resp:
            body = StreamingBody(resp)
            dl_path = Path(directory, filename or body.name)
            dl_path.parent.mkdir(exist_ok=True, parents=True)
            await body.write(dl_path,
                             overwrite=overwrite,
                             progress_bar=progress_bar)
        return dl_path

    async def download_order(self,
                             order_id: str,
                             directory: Path = Path('.'),
                             overwrite: bool = False,
                             progress_bar: bool = False) -> List[Path]:
        """Download all assets in an order.

        Parameters:
            order_id: The ID of the order.
            directory: Base directory for file download. This directory must
                already exist.
            overwrite: Overwrite files if they already exist.
            progress_bar: Show progress bar during download.

        Returns:
            Paths to downloaded files.

        Raises:
            planet.exceptions.APIError: On API error.
            planet.exceptions.ClientError: If the order is not in a final
                state.
        """
        order = await self.get_order(order_id)
        order_state = order['state']
        if not OrderStates.is_final(order_state):
            raise exceptions.ClientError(
                'Order cannot be downloaded because the order state '
                f'({order_state}) is not a final state. '
                'Consider using wait functionality before '
                'attempting to download.')

        info = self._get_download_info(order)
        LOGGER.info(f'downloading {len(info)} assets from order {order_id}')

        filenames = [
            await self.download_asset(i['location'],
                                      filename=i['filename'],
                                      directory=directory / i['directory'],
                                      overwrite=overwrite,
                                      progress_bar=progress_bar) for i in info
        ]

        return filenames

    @staticmethod
    def _get_download_info(order):
        links = order['_links']
        results = links.get('results', [])

        def _info_from_result(result):
            name = Path(result['name'])
            return {
                'location': result['location'],
                'directory': name.parent,
                'filename': name.name
            }

        try:
            info = list(_info_from_result(r) for r in results if r)
        except TypeError:
            LOGGER.warning(
                'order does not have any locations, will not download any ' +
                'files.')
            info = []
        return info

    @staticmethod
    def validate_checksum(directory: Path, checksum: str):
        """Validate checksums of downloaded files against order manifest.

        For each file entry in the order manifest, the specified checksum given
        in the manifest file will be validated against the checksum calculated
        from the downloaded file.

        Parameters:
            directory: Path to order directory.
            checksum: The type of checksum hash- 'MD5' or 'SHA256'.

        Raises:
            planet.exceptions.ClientError: If a file is missing or if checksums
                do not match.
        """
        manifest_path = directory / 'manifest.json'

        try:
            manifest_data = json.loads(manifest_path.read_text())
        except FileNotFoundError:
            raise exceptions.ClientError(
                f'File ({manifest_path}) does not exist.')
        except json.decoder.JSONDecodeError:
            raise exceptions.ClientError(
                f'Manifest file ({manifest_path}) does not contain valid JSON.'
            )

        if checksum.upper() == 'MD5':
            hash_type = hashlib.md5
        elif checksum.upper() == 'SHA256':
            hash_type = hashlib.sha256
        else:
            raise exceptions.ClientError(
                f'Checksum ({checksum}) must be one of MD5 or SHA256.')

        for json_entry in manifest_data['files']:
            origin_hash = json_entry['digests'][checksum.lower()]

            filename = directory / json_entry['path']
            try:
                returned_hash = hash_type(filename.read_bytes()).hexdigest()
            except FileNotFoundError:
                raise exceptions.ClientError(
                    f'Checksum failed. File ({filename}) does not exist.')

            if origin_hash != returned_hash:
                raise exceptions.ClientError(
                    f'File ({filename}) checksums do not match.')

    async def wait(self,
                   order_id: str,
                   state: Optional[str] = None,
                   delay: int = 5,
                   max_attempts: int = 200,
                   callback: Optional[Callable[[str], None]] = None) -> str:
        """Wait until order reaches desired state.

        Returns the state of the order on the last poll.

        This function polls the Orders API to determine the order state, with
        the specified delay between each polling attempt, until the
        order reaches a final state, or earlier state, if specified.
        If the maximum number of attempts is reached before polling is
        complete, an exception is raised. Setting 'max_attempts' to zero will
        result in no limit on the number of attempts.

        Setting 'delay' to zero results in no delay between polling attempts.
        This will likely result in throttling by the Orders API, which has
        a rate limit of 10 requests per second. If many orders are being
        polled asynchronously, consider increasing the delay to avoid
        throttling.

        By default, polling completes when the order reaches a final state.
        If 'state' is given, polling will complete when the specified earlier
        state is reached or passed.

        Example:
            ```python
            from planet.reporting import StateBar

            with StateBar() as bar:
                await wait(order_id, callback=bar.update_state)
            ```

        Parameters:
            order_id: The ID of the order.
            state: State prior to a final state that will end polling.
            delay: Time (in seconds) between polls.
            max_attempts: Maximum number of polls. Set to zero for no limit.
            callback: Function that handles state progress updates.

        Returns
            State of the order.

        Raises:
            planet.exceptions.APIError: On API error.
            planet.exceptions.ClientError: If order_id or state is not valid or
                if the maximum number of attempts is reached before the
                specified state or a final state is reached.
        """
        if state and state not in ORDER_STATE_SEQUENCE:
            raise exceptions.ClientError(
                f'{state} must be one of {ORDER_STATE_SEQUENCE}')

        # loop without end if max_attempts is zero
        # otherwise, loop until num_attempts reaches max_attempts
        num_attempts = 0
        current_state = "UNKNOWN"
        while not max_attempts or num_attempts < max_attempts:
            t = time.time()

            order = await self.get_order(order_id)
            current_state = order['state']

            LOGGER.debug(current_state)

            if callback:
                callback(current_state)

            if OrderStates.is_final(current_state) or \
                    (state and OrderStates.reached(state, current_state)):
                break

            sleep_time = max(delay - (time.time() - t), 0)
            LOGGER.debug(f'sleeping {sleep_time}s')
            await asyncio.sleep(sleep_time)

            num_attempts += 1

        if max_attempts and num_attempts >= max_attempts:
            raise exceptions.ClientError(
                f'Maximum number of attempts ({max_attempts}) reached.')

        return current_state

    async def list_orders(
            self,
            state: Optional[str] = None,
            limit: int = 100,
            source_type: Optional[str] = None,
            name: Optional[str] = None,
            name__contains: Optional[str] = None,
            created_on: Optional[str] = None,
            last_modified: Optional[str] = None,
            hosting: Optional[bool] = None,
            sort_by: Optional[str] = None) -> AsyncIterator[dict]:
        """Iterate over the list of stored orders.

        By default, order descriptions are sorted by creation date with the last created
        order returned first.

        Note:
            The name of this method is based on the API's method name. This
            method provides iteration over results, it does not get a
            single result description or return a list of descriptions.

        Parameters:
            state (str): filter by state.
            source_type (str): filter by source type.
            name (str): filter by name.
            name__contains (str): only include orders with names containing this string.
            created_on (str): filter by creation date-time or interval.
            last_modified (str): filter by last modified date-time or interval.
            hosting (bool): only return orders that contain a hosting block
                (e.g. SentinelHub hosting).
            sort_by (str): fields to sort orders by. Multiple fields can be specified,
                separated by commas. The sort direction can be specified by appending
                ' ASC' or ' DESC' to the field name. The default sort direction is
                ascending. When multiple fields are specified, the sort order is applied
                in the order the fields are listed.

                Supported fields: name, created_on, state, last_modified

                Examples:
                 * "name"
                 * "name DESC"
                 * "name,state DESC,last_modified"
            limit (int): maximum number of results to return. When set to 0, no
                maximum is applied.

        Datetime args (created_on and last_modified) can either be a date-time or an
        interval, open or closed. Date and time expressions adhere to RFC 3339. Open
        intervals are expressed using double-dots.

        Examples:
            * A date-time: "2018-02-12T23:20:50Z"
            * A closed interval: "2018-02-12T00:00:00Z/2018-03-18T12:31:12Z"
            * Open intervals: "2018-02-12T00:00:00Z/.." or "../2018-03-18T12:31:12Z"

        Yields:
            Description of an order.

        Raises:
            planet.exceptions.APIError: On API error.
            planet.exceptions.ClientError: If state is not valid.
        """
        url = self._orders_url()
        params: Dict[str, Union[str, Sequence[str], bool]] = {}
        if source_type is not None:
            params["source_type"] = source_type
        else:
            params["source_type"] = "all"
        if name is not None:
            params["name"] = name
        if name__contains is not None:
            params["name__contains"] = name__contains
        if created_on is not None:
            params["created_on"] = created_on
        if last_modified is not None:
            params["last_modified"] = last_modified
        if hosting is not None:
            params["hosting"] = hosting
        if sort_by is not None:
            params["sort_by"] = sort_by
        if state:
            if state not in ORDER_STATE_SEQUENCE:
                raise exceptions.ClientError(
                    f'Order state ({state}) is not a valid state. '
                    f'Valid states are {ORDER_STATE_SEQUENCE}')
            params['state'] = state

        response = await self._session.request(method='GET',
                                               url=url,
                                               params=params)
        async for o in Orders(response, self._session.request, limit=limit):
            yield o



# Copyright 2020 Planet Labs, Inc.
# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from datetime import datetime, timedelta, timezone
import logging
import pytest

from planet import data_filter, exceptions

LOGGER = logging.getLogger(__name__)


def test_and_filter():
    f1 = {'type': 'testf1'}
    f2 = {'type': 'testf2'}
    res = data_filter.and_filter([f1, f2])
    expected = {'type': 'AndFilter', 'config': [f1, f2]}
    assert res == expected


def test_or_filter():
    f1 = {'type': 'testf1'}
    f2 = {'type': 'testf2'}
    res = data_filter.or_filter([f1, f2])
    expected = {'type': 'OrFilter', 'config': [f1, f2]}
    assert res == expected


def test_not_filter():
    f1 = {'type': 'testf1'}
    res = data_filter.not_filter(f1)
    expected = {'type': 'NotFilter', 'config': f1}
    assert res == expected


def test__range_filter_success():

    def _test_callback(x):
        return x + 'a'

    res = data_filter._range_filter('testfilter',
                                    'testfield',
                                    gt='a',
                                    gte=None,
                                    lt='b',
                                    lte='c',
                                    callback=_test_callback)
    expected = {
        'type': 'testfilter',
        'field_name': 'testfield',
        'config': {
            'gt': 'aa', 'lt': 'ba', 'lte': 'ca'
        }
    }
    assert expected == res


def test__range_filter_nocallback():
    res = data_filter._range_filter('testfilter',
                                    'testfield',
                                    gt='a',
                                    gte=None,
                                    lt='b',
                                    lte='c')
    expected = {
        'type': 'testfilter',
        'field_name': 'testfield',
        'config': {
            'gt': 'a', 'lt': 'b', 'lte': 'c'
        }
    }
    assert expected == res


def test__range_filter_no_conditionals():

    def _test_callback(x):  # pragma: no cover
        return x + 'a'

    with pytest.raises(exceptions.PlanetError):
        data_filter._range_filter('testfilter',
                                  'testfield',
                                  gt=None,
                                  gte=None,
                                  lt=None,
                                  lte=None,
                                  callback=_test_callback)


@pytest.mark.parametrize(
    "dtime,expected",
    [(datetime(2022, 5, 1, 1, 0, 0, 1), '2022-05-01T01:00:00.000001Z'),
     (datetime(2022, 5, 1, 1, 0, 1), '2022-05-01T01:00:01Z'),
     (datetime(2022, 6, 1, 1, 1), '2022-06-01T01:01:00Z'),
     (datetime(2022, 6, 1, 1), '2022-06-01T01:00:00Z'),
     (datetime(2022, 6, 1, 1, tzinfo=timezone(timedelta(hours=1))),
      '2022-06-01T01:00:00+01:00'),
     (datetime(2022, 6, 1, 1, tzinfo=timezone(timedelta(0))),
      '2022-06-01T01:00:00+00:00')])
def test__datetime_to_rfc3339_basic(dtime, expected):
    assert data_filter._datetime_to_rfc3339(dtime) == expected


def test_date_range_filter_success():
    res = data_filter.date_range_filter('testfield',
                                        gt=datetime(2022, 6, 1, 1),
                                        lt=datetime(2022, 7, 1, 1))
    expected = {
        'type': 'DateRangeFilter',
        'field_name': 'testfield',
        'config': {
            'gt': '2022-06-01T01:00:00Z', 'lt': '2022-07-01T01:00:00Z'
        }
    }
    assert res == expected


def test_range_filter_noconditionals():
    with pytest.raises(exceptions.PlanetError):
        data_filter.range_filter('acquired')


def test_range_filter_success():
    res = data_filter.range_filter('testfield', gt=0.1, lt=0.9)
    expected = {
        'type': 'RangeFilter',
        'field_name': 'testfield',
        'config': {
            'gt': 0.1, 'lt': 0.9
        }
    }
    assert res == expected


def test_date_range_filter_noconditionals():
    with pytest.raises(exceptions.PlanetError):
        data_filter.date_range_filter('acquired')


def test_update_filter_success():
    res = data_filter.update_filter('testfield', gt=datetime(2022, 6, 1, 1))
    expected = {
        'type': 'UpdateFilter',
        'field_name': 'testfield',
        'config': {
            'gt': '2022-06-01T01:00:00Z'
        }
    }
    assert res == expected


def test_update_filter_noconditionals():
    with pytest.raises(exceptions.PlanetError):
        data_filter.update_filter('acquired')


@pytest.mark.parametrize("geom_fixture",
                         [('geom_geojson'), ('feature_geojson'),
                          ('featurecollection_geojson')])
def test_geometry_filter(geom_fixture, request, geom_geojson):
    geom = request.getfixturevalue(geom_fixture)
    res = data_filter.geometry_filter(geom)
    expected = {
        'type': 'GeometryFilter',
        'field_name': 'geometry',
        'config': geom_geojson
    }
    assert res == expected


def test_number_in_filter():
    res = data_filter.number_in_filter('testfield', [3, 3])
    expected = {
        'type': 'NumberInFilter', 'field_name': 'testfield', 'config': [3, 3]
    }
    assert res == expected


def test_string_in_filter():
    res = data_filter.string_in_filter('testfield', ['three', 'three'])
    expected = {
        'type': 'StringInFilter',
        'field_name': 'testfield',
        'config': ['three', 'three']
    }
    assert res == expected


def test_asset_filter():
    res = data_filter.asset_filter(['asset1', 'asset2'])
    expected = {'type': 'AssetFilter', 'config': ['asset1', 'asset2']}
    assert res == expected


def test_permission_filter():
    res = data_filter.permission_filter()
    expected = {'type': 'PermissionFilter', 'config': ['assets:download']}
    assert res == expected


def test_std_quality_filter():
    res = data_filter.std_quality_filter()
    expected = {
        'type': 'StringInFilter',
        'field_name': 'quality_category',
        'config': ['standard']
    }
    assert res == expected



# Copyright 2015 Planet Labs, Inc.
# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Manage data for requests and responses."""
import logging
import mimetypes
from pathlib import Path
import random
import re
import string
from typing import AsyncGenerator, Callable, List, Optional, Protocol, Union, runtime_checkable
from urllib.parse import urlparse

import httpx
from tqdm.asyncio import tqdm

from .exceptions import PagingError

LOGGER = logging.getLogger(__name__)


class Response:
    """Handles the Planet server's response to a HTTP request."""

    def __init__(self, http_response: httpx.Response):
        """Initialize object.

        Parameters:
            http_response: Response that was received from the server.
        """
        self._http_response = http_response

    def __repr__(self):
        return f'<models.Response [{self.status_code}]>'

    @property
    def status_code(self) -> int:
        """HTTP status code"""
        return self._http_response.status_code

    def json(self) -> dict:
        """Response json"""
        return self._http_response.json()


class StreamingResponse(Response):

    @property
    def headers(self) -> httpx.Headers:
        return self._http_response.headers

    @property
    def url(self) -> str:
        return str(self._http_response.url)

    @property
    def num_bytes_downloaded(self) -> int:
        return self._http_response.num_bytes_downloaded

    async def aiter_bytes(self):
        async for c in self._http_response.aiter_bytes():
            yield c

    def aiter_lines(self):
        return self._http_response.aiter_lines()

    async def aclose(self):
        await self._http_response.aclose()


class StreamingBody:
    """A representation of a streaming resource from the API."""

    def __init__(self, response: StreamingResponse):
        """Initialize the object.

        Parameters:
            response: Response that was received from the server.
        """
        self._response = response

    @property
    def name(self) -> str:
        """The name of this resource.

        The default is to use the content-disposition header value from the
        response. If not found, falls back to resolving the name from the url
        or generating a random name with the type from the response.
        """
        name = (_get_filename_from_headers(self._response.headers)
                or _get_filename_from_url(self._response.url)
                or _get_random_filename(
                    self._response.headers.get('content-type')))
        return name

    @property
    def size(self) -> int:
        """The size of the body."""
        return int(self._response.headers['Content-Length'])

    async def write(self,
                    filename: Path,
                    overwrite: bool = True,
                    progress_bar: bool = True):
        """Write the body to a file.
        Parameters:
            filename: Name to assign to downloaded file.
            overwrite: Overwrite any existing files.
            progress_bar: Show progress bar during download.
        """

        class _LOG:

            def __init__(self, total, unit, filename, disable):
                self.total = total
                self.unit = unit
                self.disable = disable
                self.previous = 0
                self.filename = str(filename)

                if not self.disable:
                    LOGGER.debug(f'writing to {self.filename}')

            def update(self, new):
                if new - self.previous > self.unit and not self.disable:
                    # LOGGER.debug(f'{new-self.previous}')
                    perc = int(100 * new / self.total)
                    LOGGER.debug(f'{self.filename}: '
                                 f'wrote {perc}% of {self.total}')
                    self.previous = new

        unit = 1024 * 1024

        mode = 'wb' if overwrite else 'xb'
        try:
            with open(filename, mode) as fp:
                _log = _LOG(self.size,
                            16 * unit,
                            filename,
                            disable=progress_bar)
                with tqdm(total=self.size,
                          unit_scale=True,
                          unit_divisor=unit,
                          unit='B',
                          desc=str(filename),
                          disable=not progress_bar) as progress:
                    previous = self._response.num_bytes_downloaded
                    async for chunk in self._response.aiter_bytes():
                        fp.write(chunk)
                        new = self._response.num_bytes_downloaded
                        _log.update(new)
                        progress.update(new - previous)
                        previous = new
        except FileExistsError:
            LOGGER.info(f'File {filename} exists, not overwriting')


def _get_filename_from_headers(headers):
    """Get a filename from the Content-Disposition header, if available.

    :param headers dict: a ``dict`` of response headers
    :returns: a filename (i.e. ``basename``)
    :rtype: str or None
    """
    cd = headers.get('content-disposition', '')
    match = re.search('filename="?([^"]+)"?', cd)
    return match.group(1) if match else None


def _get_filename_from_url(url: str) -> Optional[str]:
    """Get a filename from a url.

    Getting a name for Landsat imagery uses this function.
    """
    path = urlparse(url).path
    name = path[path.rfind('/') + 1:]
    return name or None


def _get_random_filename(content_type=None):
    """Get a pseudo-random, Planet-looking filename.

    :returns: a filename (i.e. ``basename``)
    :rtype: str
    """
    extension = mimetypes.guess_extension(content_type or '') or ''
    characters = string.ascii_letters + '0123456789'
    letters = ''.join(random.sample(characters, 8))
    name = 'planet-{}{}'.format(letters, extension)
    return name


class Paged:
    """Asynchronous iterator over results in a paged resource.

    Each returned result is a JSON dict.
    """
    LINKS_KEY = '_links'
    NEXT_KEY = 'next'
    ITEMS_KEY = 'items'

    def __init__(self,
                 response: Response,
                 request_fcn: Callable,
                 limit: int = 0):
        """
        Parameters:
            request: Request to send to server for first page.
            request_fcn: Function for submitting a request and retrieving a
            result. Must take in url and method parameters.
            limit: Maximum number of results to return. When set to 0, no
                maximum is applied.
        """
        self._request_fcn = request_fcn

        self._pages = self._get_pages(response)

        self._items: List[dict] = []

        self.i = 0
        self.limit = limit

    def __aiter__(self):
        return self

    async def __anext__(self) -> dict:
        # This was implemented because traversing _get_pages()
        # in an async generator was resulting in retrieving all the
        # pages, when the goal is to stop retrieval when the limit
        # is reached
        if self.limit and self.i >= self.limit:
            raise StopAsyncIteration

        try:
            item = self._items.pop(0)
            self.i += 1
        except IndexError:
            page = await self._pages.__anext__()
            self._items = page[self.ITEMS_KEY]
            try:
                item = self._items.pop(0)
                self.i += 1
            except IndexError:
                raise StopAsyncIteration

        return item

    async def _get_pages(self, response) -> AsyncGenerator:
        page = response.json()
        yield page

        next_url = self._next_link(page)
        while (next_url):
            LOGGER.debug('getting next page')
            response = await self._request_fcn(url=next_url, method='GET')
            page = response.json()

            # If the next URL is the same as the previous URL we will
            # get the same response and be stuck in a page cycle. This
            # has happened in development and could happen in the case
            # of a bug in the production API.
            prev_url = next_url
            next_url = self._next_link(page)

            if next_url == prev_url:
                raise PagingError(
                    "Page cycle detected at {!r}".format(next_url))

            yield page

    def _next_link(self, page):
        try:
            next_link = page[self.LINKS_KEY][self.NEXT_KEY]
            LOGGER.debug(f'next: {next_link}')
        except KeyError:
            LOGGER.debug('end of the pages')
            next_link = False
        return next_link


@runtime_checkable
class GeoInterface(Protocol):
    """represents a class that implements __geo_interface__"""

    @property
    def __geo_interface__(self) -> dict:
        raise NotImplementedError


class Feature(dict):

    @property
    def ref(self):
        if ref := self.get("properties", {}).get("pl:ref"):
            return ref
        raise AttributeError("Feature object does not contain a reference")


GeojsonLike = Union[Feature, dict, str]
"""
GeojsonLike is a type union that represents:
* a dict containing a valid GeoJSON Feature or Geometry
* an instance of a Planet Feature (e.g. the return value from `pl.features.get_items(collection_id)`)
* an instance of a class that implements __geo_interface__ (Shapely, GeoPandas geometries)
"""



"""Tests of the Subscriptions CLI (aka planet-subscriptions)

There are 7 subscriptions commands:

[x] planet subscriptions cancel
[x] planet subscriptions create
[x] planet subscriptions get
[x] planet subscriptions list
[x] planet subscriptions patch
[x] planet subscriptions results
[x] planet subscriptions update

TODO: tests for 3 options of the planet-subscriptions-results command.

"""
import itertools
import json
import respx

from click.testing import CliRunner
import pytest

from planet.cli import cli

from test_subscriptions_api import (api_mock,
                                    cancel_mock,
                                    create_mock,
                                    failing_api_mock,
                                    get_mock,
                                    patch_mock,
                                    res_api_mock,
                                    sub_summary_mock,
                                    summary_mock,
                                    update_mock,
                                    TEST_URL)

# CliRunner doesn't agree with empty options, so a list of option
# combinations which omit the empty options is best. For example,
# parametrizing 'limit' as '' and then executing
#
# CliRunner().invoke(cli.main, args=['subscriptions', 'list', limit]
#
# does not work.


@pytest.fixture
def invoke():

    def _invoke(extra_args, runner=None, **kwargs):
        runner = runner or CliRunner()
        args = ['subscriptions', f'--base-url={TEST_URL}'] + extra_args
        return runner.invoke(cli.main, args=args, **kwargs)

    return _invoke


@pytest.mark.parametrize(
    'options,expected_count',
    [(['--status=running'], 100), ([], 100), (['--source-type=catalog'], 100),
     (['--source-type=soil_water_content'], 0),
     (['--limit=1', '--status=running'], 1),
     (['--limit=2', '--pretty', '--status=running'], 2),
     (['--limit=1', '--status=preparing'], 0),
     ([
         '--name=test xyz',
         '--name-contains=xyz',
         '--created=2018-02-12T00:00:00Z/..',
         '--updated=../2018-03-18T12:31:12Z',
         '--start-time=2018-01-01T00:00:00Z',
         '--end-time=2022-01-01T00:00:00Z/2024-01-01T00:00:00Z',
         '--hosting=true',
         '--sort-by=name DESC'
     ],
      2)])
@api_mock
# Remember, parameters come before fixtures in the function definition.
def test_subscriptions_list_options(invoke, options, expected_count):
    """Prints the expected sequence of subscriptions."""
    # While developing it is handy to have click's command invoker
    # *not* catch exceptions, so we can use the pytest --pdb option.
    result = invoke(['list'] + options, catch_exceptions=False)
    assert result.exit_code == 0  # success.

    # For a start, counting the number of "id" strings in the output
    # tells us how many subscription JSONs were printed, pretty or not.
    assert result.output.count('"id"') == expected_count


@failing_api_mock
def test_subscriptions_create_failure(invoke):
    """An invalid subscription request fails to create a new subscription."""
    # This subscription request lacks the required "delivery" and
    # "source" members.
    sub = {'name': 'lol'}

    # The "-" argument says "read from stdin" and the input keyword
    # argument specifies what bytes go to the runner's stdin.
    result = invoke(
        ['create', '-'],
        input=json.dumps(sub),
        # Note: catch_exceptions=True (the default) is required if we want
        # to exercise the "translate_exceptions" decorator and test for
        # failure.
        catch_exceptions=True)

    assert result.exit_code == 1  # failure.
    assert "Error:" in result.output


# This subscription request has the members required by our fake API.
# It must be updated when we begin to test against a more strict
# imitation of the Planet Subscriptions API.
GOOD_SUB_REQUEST = {'name': 'lol', 'delivery': True, 'source': 'wut'}
GOOD_SUB_REQUEST_WITH_HOSTING = {
    'name': 'lol', 'source': 'wut', 'hosting': True
}


@pytest.mark.parametrize('cmd_arg, runner_input',
                         [('-', json.dumps(GOOD_SUB_REQUEST)),
                          (json.dumps(GOOD_SUB_REQUEST), None),
                          ('-', json.dumps(GOOD_SUB_REQUEST_WITH_HOSTING)),
                          (json.dumps(GOOD_SUB_REQUEST_WITH_HOSTING), None)])
@create_mock
def test_subscriptions_create_success(invoke, cmd_arg, runner_input):
    """Subscriptions creation succeeds with a valid subscription request."""

    # The "-" argument says "read from stdin" and the input keyword
    # argument specifies what bytes go to the runner's stdin.
    result = invoke(
        ['create', cmd_arg],
        input=runner_input,
        # Note: catch_exceptions=True (the default) is required if we want
        # to exercise the "translate_exceptions" decorator and test for
        # failure.
        catch_exceptions=True)

    assert result.exit_code == 0  # success.


# Invalid JSON.
BAD_SUB_REQUEST = '{0: "lolwut"}'


@pytest.mark.parametrize('cmd_arg, runner_input', [('-', BAD_SUB_REQUEST),
                                                   (BAD_SUB_REQUEST, None)])
def test_subscriptions_bad_request(invoke, cmd_arg, runner_input):
    """Short circuit and print help message if request is bad."""
    # The "-" argument says "read from stdin" and the input keyword
    # argument specifies what bytes go to the runner's stdin.
    result = invoke(
        ['create', cmd_arg],
        input=runner_input,
        # Note: catch_exceptions=True (the default) is required if we want
        # to exercise the "translate_exceptions" decorator and test for
        # failure.
        catch_exceptions=True)

    assert result.exit_code == 2  # bad parameter.


@failing_api_mock
def test_subscriptions_cancel_failure(invoke):
    """Cancel command exits gracefully from an API error."""
    result = invoke(
        ['cancel', 'test'],
        # Note: catch_exceptions=True (the default) is required if we want
        # to exercise the "translate_exceptions" decorator and test for
        # failure.
        catch_exceptions=True)

    assert result.exit_code == 1  # failure.


@cancel_mock
def test_subscriptions_cancel_success(invoke):
    """Cancel command succeeds."""
    result = invoke(
        ['cancel', 'test'],
        # Note: catch_exceptions=True (the default) is required if we want
        # to exercise the "translate_exceptions" decorator and test for
        # failure.
        catch_exceptions=True)

    assert result.exit_code == 0  # success.


@failing_api_mock
def test_subscriptions_update_failure(invoke):
    """Update command exits gracefully from an API error."""
    result = invoke(
        ['update', 'test', json.dumps(GOOD_SUB_REQUEST)],
        # Note: catch_exceptions=True (the default) is required if we want
        # to exercise the "translate_exceptions" decorator and test for
        # failure.
        catch_exceptions=True)

    assert result.exit_code == 1  # failure.


@update_mock
def test_subscriptions_update_success(invoke):
    """Update command succeeds."""
    request = GOOD_SUB_REQUEST.copy()
    request['name'] = 'new_name'

    result = invoke(
        ['update', 'test', json.dumps(request)],
        # Note: catch_exceptions=True (the default) is required if we want
        # to exercise the "translate_exceptions" decorator and test for
        # failure.
        catch_exceptions=True)

    assert result.exit_code == 0  # success.
    assert json.loads(result.output)['name'] == 'new_name'


@failing_api_mock
def test_subscriptions_patch_failure(invoke):
    """Patch command exits gracefully from an API error."""
    result = invoke(
        ['patch', 'test', json.dumps(GOOD_SUB_REQUEST)],
        # Note: catch_exceptions=True (the default) is required if we want
        # to exercise the "translate_exceptions" decorator and test for
        # failure.
        catch_exceptions=True)

    assert result.exit_code == 1  # failure.


@patch_mock
def test_subscriptions_patch_success(invoke):
    """Patch command succeeds."""
    request = {'name': 'test patch'}
    result = invoke(
        ['patch', 'test', json.dumps(request)],
        # Note: catch_exceptions=True (the default) is required if we want
        # to exercise the "translate_exceptions" decorator and test for
        # failure.
        catch_exceptions=True)

    assert result.exit_code == 0  # success.
    assert json.loads(result.output)['name'] == request['name']


@failing_api_mock
def test_subscriptions_get_failure(invoke):
    """Describe command exits gracefully from an API error."""
    result = invoke(
        ['get', 'test'],
        # Note: catch_exceptions=True (the default) is required if we want
        # to exercise the "translate_exceptions" decorator and test for
        # failure.
        catch_exceptions=True)

    assert result.exit_code == 1  # failure.


@get_mock
def test_subscriptions_get_success(invoke):
    """Describe command succeeds."""
    result = invoke(
        ['get', 'test'],
        # Note: catch_exceptions=True (the default) is required if we want
        # to exercise the "translate_exceptions" decorator and test for
        # failure.
        catch_exceptions=True)

    assert result.exit_code == 0  # success.
    assert json.loads(result.output)['id'] == '42'


@failing_api_mock
def test_subscriptions_results_failure(invoke):
    """Results command exits gracefully from an API error."""
    result = invoke(
        ['results', 'test'],
        # Note: catch_exceptions=True (the default) is required if we want
        # to exercise the "translate_exceptions" decorator and test for
        # failure.
        catch_exceptions=True)

    assert result.exit_code == 1  # failure.


@pytest.mark.parametrize('options,expected_count',
                         [(['--status=created'], 100), ([], 100),
                          (['--limit=1', '--status=created'], 1),
                          (['--limit=2', '--pretty', '--status=created'], 2),
                          (['--limit=1', '--status=queued'], 0)])
@res_api_mock
# Remember, parameters come before fixtures in the function definition.
def test_subscriptions_results_success(invoke, options, expected_count):
    """Describe command succeeds."""
    result = invoke(
        ['results', 'test'] + options,
        # Note: catch_exceptions=True (the default) is required if we want
        # to exercise the "translate_exceptions" decorator and test for
        # failure.
        catch_exceptions=True)

    assert result.exit_code == 0  # success.
    assert result.output.count('"id"') == expected_count


def test_request_base_success(invoke, geom_geojson):
    """Request command succeeds."""
    source = json.dumps({
        "parameters": {
            "geometry": geom_geojson,
            "start_time": "2021-03-01T00:00:00Z",
            "end_time": "2023-11-01T00:00:00Z",
            "rrule": "FREQ=MONTHLY;BYMONTH=3,4,5,6,7,8,9,10",
            "item_types": ["PSScene"],
            "asset_types": ["ortho_analytic_4b"]
        }
    })
    delivery = json.dumps({
        "type": "amazon_s3",
        "parameters": {
            "aws_access_key_id": "keyid",
            "aws_secret_access_key": "accesskey",
            "bucket": "bucket",
            "aws_region": "region"
        }
    })

    result = invoke([
        'request',
        '--name=test',
        f'--source={source}',
        f'--delivery={delivery}'
    ])

    assert source in result.output
    assert result.exit_code == 0  # success.


@pytest.mark.parametrize("geom_fixture",
                         [('geom_geojson'), ('geom_reference'),
                          ("str_geom_reference")])
def test_request_base_clip_to_source(geom_fixture, request, invoke):
    """Clip to source using command line option."""
    geom = request.getfixturevalue(geom_fixture)
    source = json.dumps({
        "parameters": {
            "geometry": geom,
            "start_time": "2021-03-01T00:00:00Z",
            "item_types": ["PSScene"],
            "asset_types": ["ortho_analytic_4b"],
        },
    })
    result = invoke([
        'request',
        '--name=test',
        f'--source={source}',
        '--delivery={"type": "fake"}',
        '--clip-to-source'
    ])

    assert result.exit_code == 0  # success.
    req = json.loads(result.output)
    tool = req["tools"][0]
    assert tool["type"] == "clip"
    assert tool["parameters"]["aoi"] == geom


def test_request_catalog_success(mock_bundles, invoke, geom_geojson):
    """Request-catalog command succeeds"""
    source = {
        "parameters": {
            "geometry": geom_geojson,
            "start_time": "2021-03-01T00:00:00Z",
            "item_types": ["PSScene"],
            "asset_types": ["ortho_analytic_4b"]
        }
    }

    result = invoke([
        'request-catalog',
        '--item-types=PSScene',
        '--asset-types=ortho_analytic_4b',
        f"--geometry={json.dumps(geom_geojson)}",
        '--start-time=2021-03-01T00:00:00'
    ])
    assert json.loads(result.output) == source
    assert result.exit_code == 0  # success.


@res_api_mock
def test_subscriptions_results_csv(invoke):
    """Get results as CSV."""
    result = invoke(["results", "test", "--csv"])
    assert result.exit_code == 0  # success.
    assert result.output.splitlines() == ["id,status", "1234-abcd,SUCCESS"]


@pytest.mark.parametrize("geom, source_type",
                         [("geom_geojson", "biomass_proxy"),
                          ("geom_reference", None),
                          ("str_geom_reference", None)])
def test_request_pv_success(invoke, geom, source_type, request):
    """Request-pv command succeeds"""
    geom = request.getfixturevalue(geom)
    if isinstance(geom, dict):
        geom = json.dumps(geom)
    cmd = [
        "request-pv",
        "--var-id=BIOMASS-PROXY_V3.0_10",
        f"--geometry={geom}",
        "--start-time=2021-03-01T00:00:00",
    ]

    if source_type:
        cmd.append(f"--var-type={source_type}")

    result = invoke(cmd)

    assert result.exit_code == 0  # success.
    source = json.loads(result.output)
    if source_type:
        assert source["type"] == "biomass_proxy"
    else:
        assert "type" not in source
    assert source["parameters"]["id"] == "BIOMASS-PROXY_V3.0_10"


@pytest.mark.parametrize(
    # Test all the combinations of the three options plus some with dupes.
    "publishing_stages",
    list(
        itertools.chain.from_iterable(
            itertools.combinations(["preview", "standard", "finalized"], i)
            for i in range(1, 4))) + [("preview", "preview"),
                                      ("preview", "finalized", "preview")])
def test_catalog_source_publishing_stages(mock_bundles,
                                          invoke,
                                          geom_geojson,
                                          publishing_stages):
    """Catalog source publishing stages are configured."""
    result = invoke([
        'request-catalog',
        '--item-types=PSScene',
        '--asset-types=ortho_analytic_4b',
        f"--geometry={json.dumps(geom_geojson)}",
        '--start-time=2021-03-01T00:00:00',
    ] + [f'--publishing-stage={stage}' for stage in publishing_stages])

    assert result.exit_code == 0  # success.
    req = json.loads(result.output)
    assert req['parameters']['publishing_stages'] == list(
        set(publishing_stages))


@pytest.mark.parametrize("time_range_type", ["acquired", "published"])
def test_catalog_source_time_range_type(mock_bundles,
                                        invoke,
                                        geom_geojson,
                                        time_range_type):
    """Catalog source time range type is configured."""
    result = invoke([
        'request-catalog',
        '--item-types=PSScene',
        '--asset-types=ortho_analytic_4b',
        f"--geometry={json.dumps(geom_geojson)}",
        '--start-time=2021-03-01T00:00:00',
        f'--time-range-type={time_range_type}',
    ])

    assert result.exit_code == 0  # success.
    req = json.loads(result.output)
    assert req['parameters']['time_range_type'] == time_range_type


@pytest.mark.parametrize(
    "hosting_option, collection_id_option, configuration_option, expected_success",
    [
        ("--hosting=sentinel_hub", None, None, True),
        ("--hosting=sentinel_hub",
         "--collection-id=7ff105c4-e0de-4910-96db-8345d86ab734",
         None,
         True),
        ("--hosting=sentinel_hub", None, "--create-configuration", True),
        ("--hosting=sentinel_hub",
         "--collection-id=7ff105c4-e0de-4910-96db-8345d86ab734",
         "--create-configuration",
         True),
    ])
def test_request_hosting(invoke,
                         geom_geojson,
                         hosting_option,
                         collection_id_option,
                         configuration_option,
                         expected_success):
    """Test request command with various hosting and collection ID options."""
    source = json.dumps({
        "parameters": {
            "geometry": geom_geojson,
            "start_time": "2021-03-01T00:00:00Z",
            "end_time": "2023-11-01T00:00:00Z",
            "rrule": "FREQ=MONTHLY;BYMONTH=3,4,5,6,7,8,9,10",
            "item_types": ["PSScene"],
            "asset_types": ["ortho_analytic_4b"]
        }
    })

    cmd = [
        'request',
        '--name=test',
        f'--source={source}',
        hosting_option,
    ]

    if collection_id_option:
        cmd.append(collection_id_option)
    if configuration_option:
        cmd.append(configuration_option)

    result = invoke(cmd)

    assert result.exit_code == 0, "Expected command to succeed."


@respx.mock
def test_item_types(invoke, mock_bundles):

    result = invoke(['item-types'])

    expected_item_types = ["SkySatScene", "SkySatCollect", "PSScene"]
    for item_type in expected_item_types:
        assert item_type in result.output
    assert result.exit_code == 0


@summary_mock
def test_summarize_subs(invoke):
    result = invoke(['summarize'])

    assert result.exit_code == 0  # success.
    summary = json.loads(result.output)
    assert summary == {
        "subscriptions": {
            "preparing": 0,
            "pending": 0,
            "running": 0,
            "completed": 0,
            "cancelled": 0,
            "suspended": 0,
            "failed": 0
        }
    }


@sub_summary_mock
def test_summarize_results(invoke):
    result = invoke(['summarize', '--subscription-id=test'])

    assert result.exit_code == 0  # success.
    summary = json.loads(result.output)
    assert summary == {
        "results": {
            "created": 0,
            "queued": 0,
            "processing": 0,
            "failed": 0,
            "cancelled": 0,
            "success": 0
        },
        "subscription": {
            "status": "pending"
        }
    }



# Copyright 2025 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

from typing import Iterator, Optional, Union
from planet.clients.features import FeaturesClient
from planet.http import Session
from planet.models import Feature, GeoInterface


class FeaturesAPI:

    _client: FeaturesClient

    def __init__(self, session: Session, base_url: Optional[str] = None):
        """
        Parameters:
            session: Open session connected to server.
            base_url: The base URL to use. Defaults to production Features API
                base url.
        """
        self._client = FeaturesClient(session, base_url)

    def list_collections(self, limit: int = 0) -> Iterator[dict]:
        """
        List the feature collections you have access to.

        Example:

        ```
        pl = Planet()
        collections = pl.features.list_collections()
        for collection in collections:
            print(collection)
        ```
        """
        collections = self._client.list_collections(limit=limit)

        try:
            while True:
                yield self._client._call_sync(collections.__anext__())
        except StopAsyncIteration:
            pass

    def get_collection(self, collection_id: str) -> dict:
        """
        Get collection metadata.

        note: when looking up or adding features to a collection,
        you can pass the collection ID directly to the add_features
        or list_features methods. It is not necessary to look up
        the collection metadata first.
        """
        collection = self._client.get_collection(collection_id)
        return self._client._call_sync(collection)

    def create_collection(self,
                          title: str,
                          description: Optional[str] = None) -> str:
        """
        Create a new collection with the given title and description, returning the collection id.

        Example:

        ```
        pl = Planet()
        collection_id = pl.features.create_collection(
          title="My Collection",
          description="a test collection"
        )
        ```
        """
        collection = self._client.create_collection(title, description)
        return self._client._call_sync(collection)

    def list_items(self,
                   collection_id: str,
                   limit: int = 0) -> Iterator[Feature]:
        """
        List features in `collection_id`.

        Returns an iterator of Features, which are `dict` instances with a
        convenience method/property for getting a feature reference (`.ref`).
        The reference can be used in the Data, Orders and Subscriptions
        APIs. Within the Python SDK, the entire Feature can generally be
        passed to functions that accept a geometry.

        example:

        ```
        pl = Planet()
        features = pl.features.list_items(collection_id)
        for feature in features:
            print(feature.ref)
            print(feature["id"])
            print(feature["geometry"])

            # use a feature directly in a search
            results = pl.data.search(["PSScene"], geometry=feature])
        ```
        """
        results = self._client.list_items(collection_id, limit=limit)

        try:
            while True:
                yield self._client._call_sync(results.__anext__())
        except StopAsyncIteration:
            pass

    def get_item(self, collection_id: str, feature_id: str) -> Feature:
        """
        Return metadata for a single feature in a collection
        """
        return self._client._call_sync(
            self._client.get_item(collection_id, feature_id))

    def add_items(self,
                  collection_id: str,
                  feature: Union[dict, GeoInterface],
                  property_id: Optional[str] = None) -> list[str]:
        """
        Add a Feature or FeatureCollection to the collection given by `collection_id`.
        Returns a list of feature references.

        collection_id: the collection to add the feature to
        feature: a dict containing a geojson Feature or FeatureCollection, or an
          instance of a class that implements __geo_interface__ (e.g. a Shapely or
          GeoPandas geometry object)
        property_id (optional): the name of a property in the `properties` block of
          the supplied feature(s). The value will become the feature's id.

        When the feature is added to the collection, it will be given an ID.
        The ID can be overriden by providing an `id` in the feature's properties.
        Title and description can also be set this way using the `title` and
        `description` properties.

        Example:

        ```
        new_features = pl.features.add_items(
          collection_id="my-collection",
          feature=feature,
        )
        ```

        note: if a geojson Geometry is supplied, it will be converted
        to a Feature. However, we recommend doing this conversion yourself
        so that you can set a title and description property.

        The return value is always a list of references, even if you only upload one
        feature.
        """
        uploaded_features = self._client.add_items(collection_id,
                                                   feature,
                                                   property_id)

        return self._client._call_sync(uploaded_features)



# Copyright 2020 Planet Labs, Inc.
# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
import asyncio
import json
import logging
from http import HTTPStatus
import math
from unittest.mock import patch

import httpx
import respx

import pytest

from planet import exceptions, http

TEST_URL = 'mock://fantastic.com'

LOGGER = logging.getLogger(__name__)


@pytest.fixture
def mock_response():

    def mocker(code, text='', json={"message": "nope"}):
        mock_request = httpx.Request('GET', 'url')
        return httpx.Response(status_code=code,
                              json=json,
                              request=mock_request)

    return mocker


def test_basesession__raise_for_status(mock_response):
    http.BaseSession._raise_for_status(
        mock_response(HTTPStatus.CREATED, json={}))

    with pytest.raises(exceptions.BadQuery):
        http.BaseSession._raise_for_status(
            mock_response(HTTPStatus.BAD_REQUEST, json={}))

    with pytest.raises(exceptions.TooManyRequests):
        http.BaseSession._raise_for_status(
            mock_response(HTTPStatus.TOO_MANY_REQUESTS, text='', json={}))

    with pytest.raises(exceptions.APIError):
        http.BaseSession._raise_for_status(mock_response(HTTPStatus.FORBIDDEN))

    with pytest.raises(exceptions.APIError):
        http.BaseSession._raise_for_status(
            mock_response(HTTPStatus.METHOD_NOT_ALLOWED, json={}))


@pytest.mark.anyio
async def test__Limiter_max_workers(monkeypatch):
    """Test that the worker cap is enforced.

    This test async queues up more tasks than the limiter worker cap. Each task
    registers itself as running, continues to run until an external flag
    is changed, then registers itself as done. Along with the tasks, a
    controller task is run asynchronously. This task waits for the other tasks
    to queue up, checks that the number of tasks running is equal to the
    limiter worker cap, then changes the state of the external flag, releasing
    the tasks.
    """
    max_workers = 2
    limiter = http._Limiter(rate_limit=0, max_workers=max_workers)

    # this value seems small enough to speed up the test but large enough
    # to avoid undue CPU churn
    short_wait = 0.001

    limiter.retry_interval = short_wait

    active = 0
    calls = 0
    hold_flag = True

    async def test_func():
        async with limiter:
            nonlocal active, calls, hold_flag
            active += 1
            calls += 1

            # wait until hold is released
            while hold_flag:
                await asyncio.sleep(short_wait)

            active -= 1

    async def control():
        nonlocal active, hold_flag

        # this value seems large enough to allow test functions to start
        # but small enough to not noticeably slow down tests
        await asyncio.sleep(.01)

        # confirm number of workers is capped
        assert active == max_workers

        # release hold so workers can complete
        hold_flag = False

    total_calls = 2 * max_workers  # just needs to be more than worker cap
    test_functions = [test_func() for _ in range(total_calls)]
    await asyncio.gather(*test_functions, control())
    assert calls == total_calls


@pytest.mark.anyio
async def test__Limiter_rate_limit(monkeypatch):
    """Test that the rate limit is enforced.

    This test async queues up tasks and then adjusts the time seen by
    _Limiter, checking that tasks are called according to the rate limit.
    """
    rate_limit = 5  # calls per second
    cadence = .2  # rate of 5/s -> period (cadence) of 200ms
    limiter = http._Limiter(rate_limit=rate_limit, max_workers=0)

    # this value seems small enough to speed up the test but large enough
    # to avoid undue CPU churn
    limiter.retry_interval = 0.001

    # this value seems large enough for one async function to wait for others
    # to progress to the next stage but small enough to not noticeably slow
    # down tests. It needs to be larger than limiter.retry_interval
    other_fcns_wait = 0.01

    # establish control over the time _Limiter reads
    current_time = 0
    monkeypatch.setattr(http._Limiter, '_get_now', lambda x: current_time)

    calls = 0

    async def test_func():
        async with limiter:
            nonlocal calls
            calls += 1

    async def control():
        nonlocal calls, current_time

        # on call gets made right out of the gate
        await asyncio.sleep(other_fcns_wait)
        assert calls == 1

        # we haven't reached cadence delay, no new calls should be made
        current_time = 0.9 * cadence
        await asyncio.sleep(other_fcns_wait)
        assert calls == 1

        # we passed cadence delay, a new call should have been made
        current_time = 1.1 * cadence
        await asyncio.sleep(other_fcns_wait)
        assert calls == 2

        # we skip forward 2 * cadence delay, only one more call should have
        # been made
        current_time = 3.2 * cadence
        await asyncio.sleep(other_fcns_wait)
        assert calls == 3
        LOGGER.debug('here')

        # final call should have been made
        current_time = 5 * cadence
        await asyncio.sleep(other_fcns_wait)
        assert calls == 4

    total_calls = 4  # needs to equal the highest number of calls in control()
    test_functions = [test_func() for _ in range(total_calls)]
    await asyncio.gather(*test_functions, control())


@pytest.mark.anyio
async def test_session_contextmanager():
    async with http.Session():
        pass


@respx.mock
@pytest.mark.anyio
@pytest.mark.parametrize('data', (None, {'boo': 'baa'}))
async def test_session_request_success(data):

    async with http.Session() as ps:
        resp_json = {'foo': 'bar'}
        route = respx.get(TEST_URL)
        route.return_value = httpx.Response(HTTPStatus.OK, json=resp_json)

        resp = await ps.request(method='GET', url=TEST_URL, json=data)
        assert resp.json() == resp_json

        # the proper headers are included and they have the expected values
        received_request = route.calls.last.request
        assert received_request.headers['x-planet-app'] == 'python-sdk'
        assert 'planet-client-python/' in received_request.headers[
            'user-agent']

        if data:
            assert received_request.headers[
                'content-type'] == 'application/json'
            assert json.loads(received_request.content) == data


@respx.mock
@pytest.mark.anyio
async def test_session_stream():
    async with http.Session() as ps:
        mock_resp = httpx.Response(HTTPStatus.OK, text='bubba')
        respx.get(TEST_URL).return_value = mock_resp

        async with ps.stream(method='GET', url=TEST_URL) as resp:
            chunks = [c async for c in resp.aiter_bytes()]
            assert chunks[0] == b'bubba'


@respx.mock
@pytest.mark.anyio
async def test_session_request_retry():
    """Test the retry in the Session.request method"""
    async with http.Session() as ps:
        route = respx.get(TEST_URL)
        route.side_effect = [
            httpx.Response(HTTPStatus.TOO_MANY_REQUESTS, json={}),
            httpx.Response(HTTPStatus.OK, json={})
        ]

        # let's not actually introduce a wait into the tests
        ps.max_retry_backoff = 0

        resp = await ps.request(method='GET', url=TEST_URL)
        assert resp
        assert route.call_count == 2


@respx.mock
@pytest.mark.anyio
async def test_session__retry():
    """A unit test for the _retry function"""

    async def test_func():
        # directly trigger the retry logic
        raise exceptions.TooManyRequests

    with patch('planet.http.Session._calculate_wait') as mock_wait:
        # let's not actually introduce a wait into the tests
        mock_wait.return_value = 0

        async with http.Session() as ps:
            with pytest.raises(exceptions.TooManyRequests):
                await ps._retry(test_func)

        calls = mock_wait.call_args_list
        args = [c[0] for c in calls]
        assert args == [(1, 64), (2, 64), (3, 64), (4, 64), (5, 64)]


def test__calculate_wait():
    max_retry_backoff = 20
    wait_times = [
        http.Session._calculate_wait(i + 1, max_retry_backoff)
        for i in range(5)
    ]

    # (min, max): 2**n to 2**n + 1, last entry hit threshold
    expected_times = [2, 4, 8, 16, 20]

    for wait, expected in zip(wait_times, expected_times):
        # this doesn't really test the randomness but does test exponential
        # and threshold
        assert math.floor(wait) == expected


@respx.mock
@pytest.mark.anyio
async def test_authsession_request():
    sess = http.AuthSession()
    resp_json = {'token': 'foobar'}
    mock_resp = httpx.Response(HTTPStatus.OK, json=resp_json)
    respx.get(TEST_URL).return_value = mock_resp

    resp = sess.request(method='GET', url=TEST_URL, json={'foo': 'bar'})
    assert resp.json() == resp_json


def test_authsession__raise_for_status(mock_response):
    with pytest.raises(exceptions.APIError):
        http.AuthSession._raise_for_status(
            mock_response(HTTPStatus.BAD_REQUEST, json={}))

    with pytest.raises(exceptions.APIError):
        http.AuthSession._raise_for_status(
            mock_response(HTTPStatus.UNAUTHORIZED, json={}))
